# -*- coding: utf-8 -*-

r"""The `flux_model` module contains classes for different flux models. The
class for the most generic flux model is `FluxModel`, which is an abstract base
class. It describes a mathematical function for the differential flux:

.. math::

    \frac{d^4\Phi(\alpha,\delta,E,t | \vec{p}_{\mathrm{s}})}{\mathrm{d}A
    \mathrm{d}\Omega \mathrm{d}E \mathrm{d}t}

"""

import abc
from astropy import (
    units,
)
import numpy as np
from scipy.integrate import (
    quad,
)
import scipy.special
import scipy.stats

from skyllh.core import (
    tool,
)
from skyllh.core.config import (
    Config,
    HasConfig,
)
from skyllh.core.math import (
    MathFunction,
)
from skyllh.core.model import (
    Model,
)
from skyllh.core.py import (
    classname,
    float_cast,
)
from skyllh.core.source_model import (
    IsPointlike,
)


class FluxProfile(
        MathFunction,
        HasConfig,
        metaclass=abc.ABCMeta):
    """The abstract base class for a flux profile math function.
    """
    def __init__(
            self,
            **kwargs,
    ):
        """Creates a new FluxProfile instance.
        """
        super().__init__(**kwargs)


class SpatialFluxProfile(
        FluxProfile,
        metaclass=abc.ABCMeta):
    """The abstract base class for a spatial flux profile function.
    """
    def __init__(
            self,
            angle_unit=None,
            **kwargs):
        """Creates a new SpatialFluxProfile instance.

        Parameters
        ----------
        angle_unit : instance of astropy.units.UnitBase | None
            The used unit for angles.
            If set to ``Ǹone``, the configured default angle unit for fluxes is
            used.
        """
        super().__init__(
            **kwargs)

        self.angle_unit = angle_unit

    @property
    def angle_unit(self):
        """The set unit of angle used for this spatial flux profile.
        If set to ``Ǹone`` the configured default angle unit for fluxes is used.
        """
        return self._angle_unit

    @angle_unit.setter
    def angle_unit(self, unit):
        if unit is None:
            unit = self._cfg['units']['defaults']['fluxes']['angle']
        if not isinstance(unit, units.UnitBase):
            raise TypeError(
                'The property angle_unit must be of type '
                'astropy.units.UnitBase!')
        self._angle_unit = unit

    @abc.abstractmethod
    def __call__(
            self,
            ra,
            dec,
            unit=None):
        """This method is supposed to return the spatial profile value for the
        given celestrial coordinates.

        Parameters
        ----------
        ra : float | 1d numpy ndarray of float
            The right-ascention coordinate.
        dec : float | 1d numpy ndarray of float
            The declination coordinate.
        unit : instance of astropy.units.UnitBase | None
            The unit of the given celestrial angles.
            If ``None``, the set angle unit of this SpatialFluxProfile is
            assumed.

        Returns
        -------
        values : 1D numpy ndarray
            The spatial profile values.
        """
        pass


class UnitySpatialFluxProfile(
        SpatialFluxProfile):
    """Spatial flux profile for the constant profile function 1 for any spatial
    coordinates.
    """
    def __init__(
            self,
            angle_unit=None,
            **kwargs,
    ):
        """Creates a new UnitySpatialFluxProfile instance.

        Parameters
        ----------
        angle_unit : instance of astropy.units.UnitBase | None
            The used unit for angles.
            If set to ``Ǹone``, the configured default angle unit for fluxes is
            used.
        """
        super().__init__(
            angle_unit=angle_unit,
            **kwargs)

    @property
    def math_function_str(self):
        """(read-only) The string representation of the mathematical function of
        this spatial flux profile instance.
        """
        return '1'

    def __call__(
            self,
            ra,
            dec,
            unit=None):
        """Returns 1 as numpy ndarray in same shape as ra and dec.

        Parameters
        ----------
        ra : float | 1d numpy ndarray of float
            The right-ascention coordinate.
        dec : float | 1d numpy ndarray of float
            The declination coordinate.
        unit : instance of astropy.units.UnitBase | None
            The unit of the given celestrial angles.
            By the definition of this class this argument is ignored.

        Returns
        -------
        values : 1D numpy ndarray
            1 in same shape as ra and dec.
        """
        (ra, dec) = np.atleast_1d(ra, dec)
        if ra.shape != dec.shape:
            raise ValueError(
                'The ra and dec arguments must be of the same shape!')

        return np.ones_like(ra)


class PointSpatialFluxProfile(
        SpatialFluxProfile):
    """Spatial flux profile for a delta function at the celestrical coordinate
    (ra, dec).
    """
    def __init__(
            self,
            ra,
            dec,
            angle_unit=None,
            **kwargs,
    ):
        """Creates a new spatial flux profile for a point at equatorial
        coordinate (ra, dec).

        Parameters
        ----------
        ra : float | None
            The right-ascention of the point.
            In case it is None, the evaluation of this spatial flux profile will
            return zero, unless evaluated for ra=None.
        dec : float | None
            The declination of the point.
            In case it is None, the evaluation of this spatial flux profile will
            return zero, unless evaluated for dec=None.
        angle_unit : instance of astropy.units.UnitBase | None
            The used unit for angles.
            If set to ``Ǹone``, the configured default angle unit for fluxes is
            used.
        """
        super().__init__(
            angle_unit=angle_unit,
            **kwargs)

        self.ra = ra
        self.dec = dec

        # Define the names of the parameters, which can be updated.
        self.param_names = ('ra', 'dec')

    @property
    def ra(self):
        """The right-ascention of the point.
        The unit is the set angle unit of this SpatialFluxProfile instance.
        """
        return self._ra

    @ra.setter
    def ra(self, v):
        v = float_cast(
            v,
            'The ra property must be castable to type float!',
            allow_None=True)
        self._ra = v

    @property
    def dec(self):
        """The declination of the point.
        The unit is the set angle unit of this SpatialFluxProfile instance.
        """
        return self._dec

    @dec.setter
    def dec(self, v):
        v = float_cast(
            v,
            'The dec property must be castable to type float!',
            allow_None=True)
        self._dec = v

    @property
    def math_function_str(self):
        """(read-only) The string representation of the mathematical function of
        this spatial flux profile instance. It is None, if the right-ascention
        or declination property is set to None.
        """
        if (self._ra is None) or (self._dec is None):
            return None

        s = (f'delta(ra-{self._ra:g}{self._angle_unit})*'
             f'delta(dec-{self._dec:g}{self._angle_unit})')

        return s

    def __call__(
            self,
            ra,
            dec,
            unit=None):
        """Returns a numpy ndarray in same shape as ra and dec with 1 if
        `ra` equals `self.ra` and `dec` equals `self.dec`, and 0 otherwise.

        Parameters
        ----------
        ra : float | 1d numpy ndarray of float
            The right-ascention coordinate at which to evaluate the spatial flux
            profile. The unit must be the internally used angle unit.
        dec : float | 1d numpy ndarray of float
            The declination coordinate at which to evaluate the spatial flux
            profile. The unit must be the internally used angle unit.
        unit : instance of astropy.units.UnitBase | None
            The unit of the given celestrial angles.
            If set to ``None``, the set angle unit of this SpatialFluxProfile
            instance is assumed.

        Returns
        -------
        value : 1D numpy ndarray of int8
            A numpy ndarray in same shape as ra and dec with 1 if `ra`
            equals `self.ra` and `dec` equals `self.dec`, and 0 otherwise.
        """
        (ra, dec) = np.atleast_1d(ra, dec)
        if ra.shape != dec.shape:
            raise ValueError(
                'The ra and dec arguments must be of the same shape!')

        if (unit is not None) and (unit != self._angle_unit):
            angle_unit_conv_factor = unit.to(self._angle_unit)
            ra = ra * angle_unit_conv_factor
            dec = dec * angle_unit_conv_factor

        value = (
            (ra == self._ra) &
            (dec == self._dec)
        ).astype(np.int8, copy=False)

        return value


class EnergyFluxProfile(
        FluxProfile,
        metaclass=abc.ABCMeta):
    """The abstract base class for an energy flux profile function.
    """
    def __init__(
            self,
            energy_unit=None,
            **kwargs):
        """Creates a new energy flux profile with a given energy unit to be used
        for flux calculation.

        Parameters
        ----------
        energy_unit : instance of astropy.units.UnitBase | None
            The used unit for energy.
            If set to ``None``, the configured default energy unit for fluxes is
            used.
        """
        super().__init__(
            **kwargs)

        # Set the energy unit.
        self.energy_unit = energy_unit

    @property
    def energy_unit(self):
        """The unit of energy used for the flux profile calculation.
        """
        return self._energy_unit

    @energy_unit.setter
    def energy_unit(self, unit):
        if unit is None:
            unit = self._cfg['units']['defaults']['fluxes']['energy']
        if not isinstance(unit, units.UnitBase):
            raise TypeError(
                'The property energy_unit must be of type '
                'astropy.units.UnitBase!')
        self._energy_unit = unit

    @abc.abstractmethod
    def __call__(
            self,
            E,
            unit=None):
        """This method is supposed to return the energy profile value for the
        given energy value.

        Parameters
        ----------
        E : float | 1d numpy ndarray of float
            The energy value for which to retrieve the energy profile value.
        unit : instance of astropy.units.UnitBase | None
            The unit of the given energy.
            If set to ``None``, the set energy unit of this EnergyFluxProfile
            is assumed.

        Returns
        -------
        values : 1D numpy ndarray of float
            The energy profile values for the given energies.
        """
        pass

    def get_integral(
            self,
            E1,
            E2,
            unit=None,
    ):
        """This is the default implementation for calculating the integral value
        of this energy flux profile in the range ``[E1, E2]``.

        .. note::

            This implementation utilizes the ``scipy.integrate.quad`` function
            to perform a generic numeric integration. Hence, this implementation
            is slow and should be reimplemented by the derived class if an
            analytic integral form is available.

        Parameters
        ----------
        E1 : float | 1d numpy ndarray of float
            The lower energy bound of the integration.
        E2 : float | 1d numpy ndarray of float
            The upper energy bound of the integration.
        unit : instance of astropy.units.UnitBase | None
            The unit of the given energies.
            If set to ``None``, the set energy unit of this EnergyFluxProfile
            instance is assumed.

        Returns
        -------
        integral : instance of ndarray
            The (n,)-shaped numpy ndarray holding the integral values of the
            given integral ranges.
        """
        E1 = np.atleast_1d(E1)
        E2 = np.atleast_1d(E2)

        if (unit is not None) and (unit != self._energy_unit):
            time_unit_conv_factor = unit.to(self._energy_unit)
            E1 = E1 * time_unit_conv_factor
            E2 = E2 * time_unit_conv_factor

        integral = np.empty((len(E1),), dtype=np.float64)

        for (i, (E1_i, E2_i)) in enumerate(zip(E1, E2)):
            integral[i] = quad(self, E1_i, E2_i, full_output=True)[0]

        return integral


class UnityEnergyFluxProfile(
        EnergyFluxProfile):
    """Energy flux profile for the constant function 1.
    """
    def __init__(
            self,
            energy_unit=None,
            **kwargs):
        """Creates a new UnityEnergyFluxProfile instance.

        Parameters
        ----------
        energy_unit : instance of astropy.units.UnitBase | None
            The used unit for energy.
            If set to ``None``, the configured default energy unit for fluxes is
            used.
        """
        super().__init__(
            energy_unit=energy_unit,
            **kwargs)

    @property
    def math_function_str(self):
        """(read-only) The string representation of the mathematical function of
        this energy flux profile.
        """
        return '1'

    def __call__(
            self,
            E,
            unit=None):
        """Returns 1 as numpy ndarray in some shape as E.

        Parameters
        ----------
        E : float | 1D numpy ndarray of float
            The energy value for which to retrieve the energy profile value.
        unit : instance of astropy.units.UnitBase | None
            The unit of the given energies.
            By definition of this specific class, this argument is ignored.

        Returns
        -------
        values : 1D numpy ndarray of int8
            1 in same shape as E.
        """
        E = np.atleast_1d(E)

        values = np.ones_like(E, dtype=np.int8)

        return values

    def get_integral(
            self,
            E1,
            E2,
            unit=None):
        """Computes the integral of this energy flux profile in the range
        [``E1``, ``E2``], which by definition is ``E2 - E1``.

        Parameters
        ----------
        E1 : float | 1d numpy ndarray of float
            The lower energy bound of the integration.
        E2 : float | 1d numpy ndarray of float
            The upper energy bound of the integration.
        unit : instance of astropy.units.UnitBase | None
            The unit of the given energies.
            If set to ``None``, the set energy unit of this EnergyFluxProfile
            instance is assumed.

        Returns
        -------
        integral : 1d ndarray of float
            The integral values of the given integral ranges.
        """
        E1 = np.atleast_1d(E1)
        E2 = np.atleast_1d(E2)

        if (unit is not None) and (unit != self._energy_unit):
            time_unit_conv_factor = unit.to(self._energy_unit)
            E1 = E1 * time_unit_conv_factor
            E2 = E2 * time_unit_conv_factor

        integral = E2 - E1

        return integral


class PowerLawEnergyFluxProfile(
        EnergyFluxProfile,
):
    r"""Energy flux profile for a power law profile with a reference energy
    ``E0`` and a spectral index ``gamma``.

    .. math::

        (E / E_0)^{-\gamma}

    """
    def __init__(
            self,
            E0,
            gamma,
            energy_unit=None,
            **kwargs):
        """Creates a new power law flux profile with the reference energy ``E0``
        and spectral index ``gamma``.

        Parameters
        ----------
        E0 : castable to float
            The reference energy.
        gamma : castable to float
            The spectral index.
        energy_unit : instance of astropy.units.UnitBase | None
            The used unit for energy.
            If set to ``None``, the configured default energy unit for fluxes is
            used.
        """
        super().__init__(
            energy_unit=energy_unit,
            **kwargs)

        self.E0 = E0
        self.gamma = gamma

        # Define the parameters which can be set via the `set_params`
        # method.
        self.param_names = ('E0', 'gamma',)

    @property
    def E0(self):
        """The reference energy in the set energy unit of this EnergyFluxProfile
        instance.
        """
        return self._E0

    @E0.setter
    def E0(self, v):
        v = float_cast(
            v,
            'Property E0 must be castable to type float!')
        self._E0 = v

    @property
    def gamma(self):
        """The spectral index.
        """
        return self._gamma

    @gamma.setter
    def gamma(self, v):
        v = float_cast(
            v,
            'Property gamma must be castable to type float!')
        self._gamma = v

    @property
    def math_function_str(self):
        """(read-only) The string representation of this energy flux profile
        instance.
        """
        s = f'(E / ({self._E0:g} {self._energy_unit}))^-{self._gamma:g}'

        return s

    def __call__(
            self,
            E,
            unit=None):
        """Returns the power law values for the given energies as numpy ndarray
        in same shape as E.

        Parameters
        ----------
        E : float | 1D numpy ndarray of float
            The energy value for which to retrieve the energy profile value.
        unit : instance of astropy.units.UnitBase | None
            The unit of the given energies.
            If set to ``None``, the set energy unit of this EnergyFluxProfile
            instance is assumed.

        Returns
        -------
        values : 1D numpy ndarray of float
            The energy profile values for the given energies.
        """
        E = np.atleast_1d(E)

        if (unit is not None) and (unit != self._energy_unit):
            E = E * unit.to(self._energy_unit)

        value = np.power(E / self._E0, -self._gamma)

        return value

    def get_integral(
            self,
            E1,
            E2,
            unit=None):
        """Computes the integral value of this power-law energy flux profile in
        the range ``[E1, E2]``.

        Parameters
        ----------
        E1 : float | 1d numpy ndarray of float
            The lower energy bound of the integration.
        E2 : float | 1d numpy ndarray of float
            The upper energy bound of the integration.
        unit : instance of astropy.units.UnitBase | None
            The unit of the given energies.
            If set to ``None``, the set energy unit of this EnergyFluxProfile
            instance is assumed.

        Returns
        -------
        integral : 1d ndarray of float
            The integral values of the given integral ranges.
        """
        E1 = np.atleast_1d(E1)
        E2 = np.atleast_1d(E2)

        if (unit is not None) and (unit != self._energy_unit):
            time_unit_conv_factor = unit.to(self._energy_unit)
            E1 = E1 * time_unit_conv_factor
            E2 = E2 * time_unit_conv_factor

        gamma = self._gamma

        # Handle special case for gamma = 1.
        if gamma == 1:
            integral = self._E0 * np.log(E2/E1)
            return integral

        integral = (
            np.power(self._E0, gamma) / (1-gamma) *
            (np.power(E2, 1-gamma) - np.power(E1, 1-gamma))
        )

        return integral


class CutoffPowerLawEnergyFluxProfile(
        PowerLawEnergyFluxProfile,
):
    r"""Cut-off power law energy flux profile of the form

    .. math::

        (E / E_0)^{-\gamma} \exp(-E/E_{\mathrm{cut}})

    """
    def __init__(
            self,
            E0,
            gamma,
            Ecut,
            energy_unit=None,
            **kwargs,
    ):
        """Creates a new cut-off power law flux profile with the reference
        energy ``E0``, spectral index ``gamma``, and cut-off energy ``Ecut``.

        Parameters
        ----------
        E0 : castable to float
            The reference energy.
        gamma : castable to float
            The spectral index.
        Ecut : castable to float
            The cut-off energy.
        energy_unit : instance of astropy.units.UnitBase | None
            The used unit for energy.
            If set to ``None``, the configured default energy unit for fluxes is
            used.
        """
        super().__init__(
            E0=E0,
            gamma=gamma,
            energy_unit=energy_unit,
            **kwargs)

        self.Ecut = Ecut

    @property
    def Ecut(self):
        """The energy cut value.
        """
        return self._Ecut

    @Ecut.setter
    def Ecut(self, v):
        v = float_cast(
            v,
            'The Property Ecut  must be castable to type float!')
        self._Ecut = v

    @property
    def math_function_str(self):
        """(read-only) The string representation of this energy flux profile
        instance.
        """
        s = (f'(E / ({self._E0:g} {self._energy_unit}))^-{self._gamma:g} '
             f'exp(-E / ({self._Ecut:g} {self._energy_unit}))')

        return s

    def __call__(
            self,
            E,
            unit=None,
    ):
        """Returns the cut-off power law values for the given energies as
        numpy ndarray in the same shape as E.

        Parameters
        ----------
        E : float | instance of numpy ndarray
            The energy value(s) for which to retrieve the energy profile value.
        unit : instance of astropy.units.UnitBase | None
            The unit of the given energies.
            If set to ``None``, the set energy unit of this EnergyFluxProfile
            instance is assumed.

        Returns
        -------
        values : instance of numpy ndarray
            The energy profile values for the given energies.
        """
        E = np.atleast_1d(E)

        if (unit is not None) and (unit != self._energy_unit):
            E = E * unit.to(self._energy_unit)

        values = super().__call__(E=E, unit=None)
        values *= np.exp(-E / self._Ecut)

        return values


class LogParabolaPowerLawEnergyFluxProfile(
        PowerLawEnergyFluxProfile,
):
    r"""This class provides an energy flux profile for a power-law with a
    spectral index that varies as a log parabola in energy of the form

    .. math::

        \frac{E}{E_0}^{-\left(\alpha + \beta\log(\frac{E}{E_0})\right)}

    """
    def __init__(
            self,
            E0,
            alpha,
            beta,
            energy_unit=None,
            **kwargs,
    ):
        """
        Parameters
        ----------
        E0 : castable to float
            The reference energy.
        alpha : float
            The alpha parameter of the log-parabola spectral index.
        beta : float
            The beta parameter of the log-parabola spectral index.
        energy_unit : instance of astropy.units.UnitBase | None
            The used unit for energy.
            If set to ``None``, the configured default energy unit for fluxes is
            used.
        """
        super().__init__(
            E0=E0,
            gamma=np.nan,
            energy_unit=energy_unit,
            **kwargs)

        self.alpha = alpha
        self.beta = beta

    @property
    def alpha(self):
        """The alpha parameter of the log-parabola spectral index.
        """
        return self._alpha

    @alpha.setter
    def alpha(self, v):
        v = float_cast(
            v,
            'Property alpha must be castable to type float!')
        self._alpha = v

    @property
    def beta(self):
        """The beta parameter of the log-parabola spectral index.
        """
        return self._beta

    @beta.setter
    def beta(self, v):
        v = float_cast(
            v,
            'Property beta must be castable to type float!')
        self._beta = v

    @property
    def math_function_str(self):
        """(read-only) The string representation of this energy flux profile
        instance.
        """
        s_E0 = f'{self._E0:g} {self._energy_unit}'
        s = (
            f'(E / {s_E0})'
            f'^(-({self._alpha:g} + {self._beta:g} log(E / {s_E0})))'
        )

        return s

    def __call__(
            self,
            E,
            unit=None,
    ):
        """Returns the log-parabola power-law values for the given energies as
        numpy ndarray in the same shape as E.

        Parameters
        ----------
        E : float | instance of numpy ndarray
            The energy value(s) for which to retrieve the energy profile value.
        unit : instance of astropy.units.UnitBase | None
            The unit of the given energies.
            If set to ``None``, the set energy unit of this EnergyFluxProfile
            instance is assumed.

        Returns
        -------
        values : instance of numpy ndarray
            The energy profile values for the given energies.
        """
        E = np.atleast_1d(E)

        if (unit is not None) and (unit != self._energy_unit):
            E = E * unit.to(self._energy_unit)

        values = np.power(
            E / self._E0,
            -self._alpha - self._beta * np.log(E / self._E0)
        )

        return values


class PhotosplineEnergyFluxProfile(
        EnergyFluxProfile,
):
    """The abstract base class for an energy flux profile based on a
    photospline.
    """
    @tool.requires('photospline')
    def __init__(
            self,
            splinetable,
            crit_log10_energy_lower,
            crit_log10_energy_upper,
            energy_unit=None,
            **kwargs,
    ):
        """Creates a new instance of PhotosplineEnergyFluxProfile.

        Parameters
        ----------
        splinetable : instance of photospline.SplineTable
            The instance of photospline.SplineTable representing the energy flux
            profile as a spline.
        crit_log10_energy_lower : float
            The lower edge of the spline's supported energy range in log10(E).
        crit_log10_energy_upper : float
            The upper edge of the spline's supported energy range in log10(E).
        energy_unit : instance of astropy.units.UnitBase | None
            The used unit for energy.
            If set to ``None``, the configured default energy unit for fluxes is
            used.
        """
        super().__init__(
            energy_unit=energy_unit,
            **kwargs)

        self.photospline = tool.get('photospline')

        self.splinetable = splinetable
        self.crit_log10_energy_lower = crit_log10_energy_lower
        self.crit_log10_energy_upper = crit_log10_energy_upper

    @property
    def splinetable(self):
        """The instance of photospline.SplineTable that describes the neutrino
        energy flux profile as function of neutrino energy via B-spline
        interpolation.
        """
        return self._splinetable

    @splinetable.setter
    def splinetable(self, table):
        if not isinstance(table, self.photospline.SplineTable):
            raise TypeError(
                'The splinetable property must be an instance of '
                'photospline.SplineTable! '
                f'Its current type is {classname(table)}!')
        self._splinetable = table

    @property
    def crit_log10_energy_lower(self):
        """The lower energy bound of the spline's support.
        """
        return self._crit_log10_energy_lower

    @crit_log10_energy_lower.setter
    def crit_log10_energy_lower(self, v):
        v = float_cast(
            v,
            'The property crit_log10_energy_lower must be castable to type '
            'float!')
        self._crit_log10_energy_lower = v

    @property
    def crit_log10_energy_upper(self):
        """The upper energy bound of the spline's support.
        """
        return self._crit_log10_energy_upper

    @crit_log10_energy_upper.setter
    def crit_log10_energy_upper(self, v):
        v = float_cast(
            v,
            'The property crit_log10_energy_upper must be castable to type '
            'float!')
        self._crit_log10_energy_upper = v


class TimeFluxProfile(
        FluxProfile,
        metaclass=abc.ABCMeta,
):
    """The abstract base class for a time flux profile function.
    """
    def __init__(
            self,
            t_start=-np.inf,
            t_stop=np.inf,
            time_unit=None,
            **kwargs):
        """Creates a new time flux profile instance.

        Parameters
        ----------
        t_start : float
            The start time of the time profile.
            If set to -inf, it means, that the profile starts at the beginning
            of the entire time-span of the dataset.
        t_stop : float
            The stop time of the time profile.
            If set to +inf, it means, that the profile ends at the end of the
            entire time-span of the dataset.
        time_unit : instance of astropy.units.UnitBase | None
            The used unit for time.
            If set to ``None``, the configured default time unit for fluxes is
            used.
        """
        super().__init__(
            **kwargs)

        self.time_unit = time_unit

        self.t_start = t_start
        self.t_stop = t_stop

        # Define the parameters which can be set via the `set_params`
        # method.
        self.param_names = ('t_start', 't_stop')

    @property
    def t_start(self):
        """The start time of the time profile. Can be -inf which means, that
        the profile starts at the beginning of the entire dataset.
        """
        return self._t_start

    @t_start.setter
    def t_start(self, t):
        t = float_cast(
            t,
            'The t_start property must be castable to type float! '
            f'Its current type is {classname(t)}!')
        self._t_start = t

    @property
    def t_stop(self):
        """The stop time of the time profile. Can be +inf which means, that
        the profile ends at the end of the entire dataset.
        """
        return self._t_stop

    @t_stop.setter
    def t_stop(self, t):
        t = float_cast(
            t,
            'The t_stop property must be castable to type float! '
            f'Its current type is {classname(t)}!')
        self._t_stop = t

    @property
    def duration(self):
        """(read-only) The duration of the time profile.
        """
        return self._t_stop - self._t_start

    @property
    def time_unit(self):
        """The unit of time used for the flux profile calculation.
        """
        return self._time_unit

    @time_unit.setter
    def time_unit(self, unit):
        if unit is None:
            unit = self._cfg['units']['defaults']['fluxes']['time']
        if not isinstance(unit, units.UnitBase):
            raise TypeError(
                'The property time_unit must be of type '
                'astropy.units.UnitBase! '
                f'Its current type is {classname(unit)}!')
        self._time_unit = unit

    def get_total_integral(self):
        """Calculates the total integral of the time profile from t_start to
        t_stop.

        Returns
        -------
        integral : float
            The integral value of the entire time profile.
            The value is in the set time unit of this TimeFluxProfile instance.
        """
        integral = self.get_integral(self._t_start, self._t_stop).squeeze()

        return integral

    @abc.abstractmethod
    def __call__(
            self,
            t,
            unit=None,
    ):
        """This method is supposed to return the time profile value for the
        given times.

        Parameters
        ----------
        t : float | 1D numpy ndarray of float
            The time(s) for which to get the time flux profile values.
        unit : instance of astropy.units.UnitBase | None
            The unit of the given times.
            If set to ``None``, the set time unit of this TimeFluxProfile
            instance is assumed.

        Returns
        -------
        values : 1D numpy ndarray of float
            The time profile values.
        """
        pass

    @abc.abstractmethod
    def move(
            self,
            dt,
            unit=None,
    ):
        """Abstract method to move the time profile by the given amount of time.

        Parameters
        ----------
        dt : float
            The time difference of how far to move the time profile in time.
            This can be a positive or negative time shift value.
        unit : instance of astropy.units.UnitBase | None
            The unit of the given time difference.
            If set to ``Ǹone``, the set time unit of this TimeFluxProfile
            instance is assumed.
        """
        pass

    @abc.abstractmethod
    def get_integral(
            self,
            t1,
            t2,
            unit=None,
    ):
        """This method is supposed to calculate the integral of the time profile
        from time ``t1`` to time ``t2``.

        Parameters
        ----------
        t1 : float | array of float
            The start time of the integration.
        t2 : float | array of float
            The end time of the integration.
        unit : instance of astropy.units.UnitBase | None
            The unit of the given times.
            If set to ``Ǹone``, the set time unit of this TimeFluxProfile
            instance is assumed.

        Returns
        -------
        integral : array of float
            The integral value(s) of the time profile. The values are in the
            set time unit of this TimeFluxProfile instance.
        """
        pass


class UnityTimeFluxProfile(
        TimeFluxProfile,
):
    """Time flux profile for the constant profile function ``1``.
    """
    def __init__(
            self,
            time_unit=None,
            **kwargs,
    ):
        """Creates a new instance of UnityTimeFluxProfile.

        Parameters
        ----------
        time_unit : instance of astropy.units.UnitBase | None
            The used unit for time.
            If set to ``None``, the configured default time unit for fluxes is
            used.
        """
        super().__init__(
            time_unit=time_unit,
            **kwargs)

    @property
    def math_function_str(self):
        return '1'

    def __call__(
            self,
            t,
            unit=None,
    ):
        """Returns 1 as numpy ndarray in same shape as t.

        Parameters
        ----------
        t : float | 1D numpy ndarray of float
            The time(s) for which to get the time flux profile values.
        unit : instance of astropy.units.UnitBase | None
            The unit of the given times.
            By definition of this specific class, this argument is ignored.

        Returns
        -------
        values : 1D numpy ndarray of int8
            1 in same shape as ``t``.
        """
        t = np.atleast_1d(t)

        values = np.ones_like(t, dtype=np.int8)

        return values

    def move(
            self,
            dt,
            unit=None,
    ):
        """Moves the time profile by the given amount of time. By definition
        this method does nothing, because the profile is 1 over the entire
        dataset time range.

        Parameters
        ----------
        dt : float
            The time difference of how far to move the time profile in time.
            This can be a positive or negative time shift value.
        unit : instance of astropy.units.UnitBase | None
            The unit of the given time difference.
            If set to ``None``, the set time unit of this TimeFluxProfile
            instance is assumed.
        """
        pass

    def get_integral(
            self,
            t1,
            t2,
            unit=None,
    ):
        """Calculates the integral of the time profile from time t1 to time t2.

        Parameters
        ----------
        t1 : float | array of float
            The start time of the integration.
        t2 : float | array of float
            The end time of the integration.
        unit : instance of astropy.units.UnitBase | None
            The unit of the given times.
            If set to ``None``, the set time unit of this TimeFluxProfile
            instance is assumed.

        Returns
        -------
        integral : array of float
            The integral value(s) of the time profile. The values are in the
            set time unit of this TimeFluxProfile instance.
        """
        if (unit is not None) and (unit != self._time_unit):
            time_unit_conv_factor = unit.to(self._time_unit)
            t1 = t1 * time_unit_conv_factor
            t2 = t2 * time_unit_conv_factor

        integral = t2 - t1

        return integral


class BoxTimeFluxProfile(
        TimeFluxProfile,
):
    """This class describes a box-shaped time flux profile.
    It has the following parameters:

        t0 : float
            The mid time of the box profile.
        tw : float
            The width of the box profile.

    The box is centered at ``t0`` and extends to +/-``tw``/2 around ``t0``.
    """

    @classmethod
    def from_start_and_stop_time(
            cls,
            start,
            stop,
            time_unit=None,
            **kwargs,
    ):
        """Constructs a BoxTimeFluxProfile instance from the given start and
        stop time.

        Parameters
        ----------
        start : float
            The start time of the box profile.
        stop : float
            The stop time of the box profile.
        time_unit : instance of astropy.units.UnitBase | None
            The used unit for time.
            If set to ``None``, the configured default time unit for fluxes is
            used.
        **kwargs
            Additional keyword arguments, which are passed to the constructor
            of the :class:`BoxTimeFluxProfile` class.

        Returns
        -------
        profile : instance of BoxTimeFluxProfile
            The newly created instance of BoxTimeFluxProfile.
        """
        t0 = 0.5*(start + stop)
        tw = stop - start

        profile = cls(
            t0=t0,
            tw=tw,
            time_unit=time_unit,
            **kwargs)

        return profile

    def __init__(
            self,
            t0,
            tw,
            time_unit=None,
            **kwargs,
    ):
        """Creates a new box-shaped time profile instance.

        Parameters
        ----------
        t0 : float
            The mid time of the box profile.
        tw : float
            The width of the box profile.
        time_unit : instance of astropy.units.UnitBase | None
            The used unit for time.
            If set to ``None``, the configured default time unit for fluxes is
            used.
        """
        t_start = t0 - tw/2.
        t_stop = t0 + tw/2.

        super().__init__(
            t_start=t_start,
            t_stop=t_stop,
            time_unit=time_unit,
            **kwargs)

        # Define the parameters which can be set via the `set_params`
        # method.
        self.param_names = ('t0', 'tw')

    @property
    def t0(self):
        """The time of the mid point of the box.
        The value is in the set time unit of this TimeFluxProfile instance.
        """
        return 0.5*(self._t_start + self._t_stop)

    @t0.setter
    def t0(self, t):
        old_t0 = self.t0
        dt = t - old_t0
        self.move(dt)

    @property
    def tw(self):
        """The time width of the box.
        The value is in the set time unit of this TimeFluxProfile instance.
        """
        return self._t_stop - self._t_start

    @tw.setter
    def tw(self, w):
        t0 = self.t0
        self._t_start = t0 - 0.5*w
        self._t_stop = t0 + 0.5*w

    @property
    def math_function_str(self):
        """The string representation of the mathematical function of this
        TimeFluxProfile instance.
        """
        t0 = self.t0
        tw = self.tw

        s = f'1 for t in [{t0:g}-{tw:g}/2; {t0:g}+{tw:g}/2], 0 otherwise'

        return s

    def __call__(
            self,
            t,
            unit=None,
    ):
        """Returns 1 for all t within the interval [t0-tw/2; t0+tw/2], and 0
        otherwise.

        Parameters
        ----------
        t : float | 1D numpy ndarray of float
            The time(s) for which to get the time flux profile values.
        unit : instance of astropy.units.UnitBase | None
            The unit of the given times.
            If set to ``None``, the set time unit of this TimeFluxProfile
            instance is assumed.

        Returns
        -------
        values : 1D numpy ndarray of int8
            The value(s) of the time flux profile for the given time(s).
        """
        t = np.atleast_1d(t)

        if (unit is not None) and (unit != self._time_unit):
            t = t * unit.to(self._time_unit)

        values = np.zeros((t.shape[0],), dtype=np.int8)
        m = (t >= self._t_start) & (t <= self._t_stop)
        values[m] = 1

        return values

    def cdf(
            self,
            t,
            unit=None,
    ):
        """Calculates the cumulative distribution function value for the given
        time values ``t``.

        Parameters
        ----------
        t : float | instance of numpy ndarray
            The (N_times,)-shaped numpy ndarray holding the time values for
            which to calculate the CDF values.
        unit : instance of astropy.units.UnitBase | None
            The unit of the given times.
            If set to ``None``, the set time unit of this TimeFluxProfile is
            assumed.

        Returns
        -------
        values : instance of numpy ndarray
            The (N_times,)-shaped numpy ndarray holding the cumulative
            distribution function values for each time ``t``.
        """
        t = np.atleast_1d(t)

        if (unit is not None) and (unit != self._time_unit):
            t = t * unit.to(self._time_unit)

        t_start = self._t_start
        t_stop = self._t_stop

        values = np.zeros(t.size, dtype=np.float64)

        m = (t_start <= t) & (t <= t_stop)
        values[m] = (t[m] - t_start) / (t_stop - t_start)

        m = (t > t_stop)
        values[m] = 1

        return values

    def move(
            self,
            dt,
            unit=None,
    ):
        """Moves the box-shaped time profile by the time difference dt.

        Parameters
        ----------
        dt : float
            The time difference of how far to move the time profile in time.
            This can be a positive or negative time shift value.
        unit : instance of astropy.units.UnitBase | None
            The unit of ``dt``.
            If set to ``None``, the set time unit of this TimeFluxProfile
            instance is assumed.
        """
        if (unit is not None) and (unit != self._time_unit):
            dt = dt * unit.to(self._time_unit)

        self._t_start += dt
        self._t_stop += dt

    def get_integral(
            self,
            t1,
            t2,
            unit=None,
    ):
        """Calculates the integral of the box-shaped time flux profile from
        time t1 to time t2.

        Parameters
        ----------
        t1 : float | array of float
            The start time(s) of the integration.
        t2 : float | array of float
            The end time(s) of the integration.
        unit : instance of astropy.units.UnitBase | None
            The unit of the given times.
            If set to ``None``, the set time unit of this TimeFluxProfile
            instance is assumed.

        Returns
        -------
        integral : array of float
            The integral value(s). The values are in the set time unit of this
            TimeFluxProfile instance.
        """
        t1 = np.atleast_1d(t1)
        t2 = np.atleast_1d(t2)

        if (unit is not None) and (unit != self._time_unit):
            time_unit_conv_factor = unit.to(self._time_unit)
            t1 = t1 * time_unit_conv_factor
            t2 = t2 * time_unit_conv_factor

        integral = np.zeros((t1.shape[0],), dtype=np.float64)

        m = (t2 >= self._t_start) & (t1 <= self._t_stop)
        N = np.count_nonzero(m)

        t1 = np.max(np.vstack((t1[m], np.repeat(self._t_start, N))).T, axis=1)
        t2 = np.min(np.vstack((t2[m], np.repeat(self._t_stop, N))).T, axis=1)

        integral[m] = t2 - t1

        return integral


class GaussianTimeFluxProfile(
        TimeFluxProfile,
):
    """This class describes a gaussian-shaped time flux profile.
    It has the following parameters:

        t0 : float
            The mid time of the gaussian profile.
        sigma_t : float
            The one-sigma width of the gaussian profile.
    """

    def __init__(
            self,
            t0,
            sigma_t,
            tol=1e-12,
            time_unit=None,
            **kwargs):
        """Creates a new gaussian-shaped time flux profile instance.

        Parameters
        ----------
        t0 : float
            The mid time of the gaussian profile.
        sigma_t : float
            The one-sigma width of the gaussian profile.
        tol : float
            The tolerance of the gaussian value. This defines the start and end
            time of the gaussian profile.
        time_unit : instance of astropy.units.UnitBase | None
            The used unit for time.
            If set to ``None``, the configured default time unit for fluxes is
            used.
        """
        # Calculate the start and end time of the gaussian profile, such that
        # at those times the gaussian values obey the given tolerance.
        dt = np.sqrt(-2 * sigma_t**2 * np.log(tol))
        t_start = t0 - dt
        t_stop = t0 + dt

        super().__init__(
            t_start=t_start,
            t_stop=t_stop,
            time_unit=time_unit,
            **kwargs)

        self.t0 = t0
        self.sigma_t = sigma_t

        # Define the parameters which can be set via the `set_params`
        # method.
        self.param_names = ('t0', 'sigma_t')

    @property
    def math_function_str(self):
        return 'exp(-(t-t0)^2/(2 sigma_t^2))'

    @property
    def t0(self):
        """The time of the mid point of the gaussian profile.
        The unit of the value is the set time unit of this TimeFluxProfile
        instance.
        """
        return 0.5*(self._t_start + self._t_stop)

    @t0.setter
    def t0(self, t):
        t = float_cast(
            t,
            'The t0 property must be castable to type float!')
        old_t0 = self.t0
        dt = t - old_t0
        self.move(dt)

    @property
    def sigma_t(self):
        """The one-sigma width of the gaussian profile.
        The unit of the value is the set time unit of this TimeFluxProfile
        instance.
        """
        return self._sigma_t

    @sigma_t.setter
    def sigma_t(self, sigma):
        sigma = float_cast(
            sigma,
            'The sigma_t property must be castable to type float!')
        self._sigma_t = sigma

    def __call__(
            self,
            t,
            unit=None,
    ):
        """Returns the gaussian profile value for the given time ``t``.

        Parameters
        ----------
        t : float | 1D numpy ndarray of float
            The time(s) for which to get the time flux profile values.
        unit : instance of astropy.units.UnitBase | None
            The unit of the given times.
            If set to ``None``, the set time unit of this TimeFluxProfile is
            assumed.

        Returns
        -------
        values : 1D numpy ndarray of float
            The value(s) of the time flux profile for the given time(s).
        """
        t = np.atleast_1d(t)

        if (unit is not None) and (unit != self._time_unit):
            time_unit_conv_factor = unit.to(self._time_unit)
            t = t * time_unit_conv_factor

        m = (t >= self.t_start) & (t < self.t_stop)

        s = self._sigma_t
        twossq = 2*s*s
        t0 = 0.5*(self._t_stop + self._t_start)
        dt = t[m] - t0

        values = np.zeros_like(t)
        values[m] = np.exp(-dt*dt/twossq)

        return values

    def cdf(
            self,
            t,
            unit=None,
    ):
        """Calculates the cumulative distribution function values for the given
        time values ``t``.

        Parameters
        ----------
        t : float | instance of numpy ndarray
            The (N_times,)-shaped numpy ndarray holding the time values for
            which to calculate the CDF values.
        unit : instance of astropy.units.UnitBase | None
            The unit of the given times.
            If set to ``None``, the set time unit of this TimeFluxProfile is
            assumed.

        Returns
        -------
        values : instance of numpy ndarray
            The (N_times,)-shaped numpy ndarray holding the cumulative
            distribution function values for each time ``t``.
        """
        t = np.atleast_1d(t)

        if (unit is not None) and (unit != self._time_unit):
            t = t * unit.to(self._time_unit)

        t_start = self._t_start
        t_stop = self._t_stop

        values = np.zeros(t.size, dtype=np.float64)

        m = (t_start <= t) & (t <= t_stop)
        values[m] = (
            self.get_integral(t1=t_start, t2=t[m]) / self.get_total_integral()
        )

        m = (t > t_stop)
        values[m] = 1

        return values

    def move(
            self,
            dt,
            unit=None,
    ):
        """Moves the gaussian time profile by the given amount of time.

        Parameters
        ----------
        dt : float
            The time difference of how far to move the time profile in time.
            This can be a positive or negative time shift value.
        unit : instance of astropy.units.UnitBase | None
            The unit of the given time difference.
            If set to ``None``, the set time unit of this TimeFluxProfile is
            assumed.
        """
        if (unit is not None) and (unit != self._time_unit):
            dt = dt * unit.to(self._time_unit)

        self._t_start += dt
        self._t_stop += dt

    def get_integral(
            self,
            t1,
            t2,
            unit=None,
    ):
        """Calculates the integral of the gaussian time profile from time ``t1``
        to time ``t2``.

        Parameters
        ----------
        t1 : float | array of float
            The start time(s) of the integration.
        t2 : float | array of float
            The end time(s) of the integration.
        unit : instance of astropy.units.UnitBase | None
            The unit of the given times.
            If set to ``None``, the set time unit of this TimeFluxProfile
            instance is assumed.

        Returns
        -------
        integral : array of float
            The integral value(s). The values are in the set time unit of
            this TimeFluxProfile instance.
        """
        if (unit is not None) and (unit != self._time_unit):
            time_unit_conv_factor = unit.to(self._time_unit)
            t1 = t1 * time_unit_conv_factor
            t2 = t2 * time_unit_conv_factor

        t0 = 0.5*(self._t_stop + self._t_start)
        sigma_t = self._sigma_t

        c1 = np.sqrt(np.pi/2) * sigma_t
        c2 = np.sqrt(2) * sigma_t
        i1 = c1 * scipy.special.erf((t1 - t0)/c2)
        i2 = c1 * scipy.special.erf((t2 - t0)/c2)

        integral = i2 - i1

        return integral


class FluxModel(
        MathFunction,
        HasConfig,
        Model,
        metaclass=abc.ABCMeta,
):
    r"""Abstract base class for all flux models of the form

    .. math::

        \Phi_S(\alpha,\delta,E,t | \vec{x}_s,\vec{p}_s).

    This base class defines the units used for the flux calculation. The unit
    of the flux is ([angle]^{-2} [energy]^{-1} [length]^{-2} [time]^{-1}).

    At this point the functional form of the flux model is not yet defined.
    """
    @staticmethod
    def get_default_units(cfg):
        """Returns the configured default units for flux models.

        Parameters
        ----------
        cfg : instance of Config
            The instance of Config holding the local configuration.

        Returns
        -------
        units_dict : dict
            The dictionary holding the configured default units used for flux
            models.
        """
        return cfg['units']['defaults']['fluxes']

    def __init__(
            self,
            angle_unit=None,
            energy_unit=None,
            length_unit=None,
            time_unit=None,
            **kwargs):
        """Creates a new FluxModel instance and defines the user-defined units.

        Parameters
        ----------
        angle_unit : instance of astropy.units.UnitBase | None
            The used unit for angles.
            If set to ``None``, the configured default angle unit for fluxes is
            used.
        energy_unit : instance of astropy.units.UnitBase | None
            The used unit for energy.
            If set to ``None``, the configured default energy unit for fluxes is
            used.
        length_unit : instance of astropy.units.UnitBase | None
            The used unit for length.
            If set to ``None``, the configured default length unit for fluxes is
            used.
        time_unit : instance of astropy.units.UnitBase | None
            The used unit for time.
            If set to ``None``, the configured default time unit for fluxes is
            used.
        """
        super().__init__(
            **kwargs)

        # Define the units.
        self.angle_unit = angle_unit
        self.energy_unit = energy_unit
        self.length_unit = length_unit
        self.time_unit = time_unit

    @property
    def angle_unit(self):
        """The unit of angle used for the flux calculation.
        """
        return self._angle_unit

    @angle_unit.setter
    def angle_unit(self, unit):
        if unit is None:
            unit = self._cfg['units']['defaults']['fluxes']['angle']
        if not isinstance(unit, units.UnitBase):
            raise TypeError(
                'The property angle_unit must be of type '
                'astropy.units.UnitBase!')
        self._angle_unit = unit

    @property
    def energy_unit(self):
        """The unit of energy used for the flux calculation.
        """
        return self._energy_unit

    @energy_unit.setter
    def energy_unit(self, unit):
        if unit is None:
            unit = self._cfg['units']['defaults']['fluxes']['energy']
        if not isinstance(unit, units.UnitBase):
            raise TypeError(
                'The property energy_unit must be of type '
                'astropy.units.UnitBase!')
        self._energy_unit = unit

    @property
    def length_unit(self):
        """The unit of length used for the flux calculation.
        """
        return self._length_unit

    @length_unit.setter
    def length_unit(self, unit):
        if unit is None:
            unit = self._cfg['units']['defaults']['fluxes']['length']
        if not isinstance(unit, units.UnitBase):
            raise TypeError(
                'The property length_unit must be of type '
                'astropy.units.UnitBase!')
        self._length_unit = unit

    @property
    def time_unit(self):
        """The unit of time used for the flux calculation.
        """
        return self._time_unit

    @time_unit.setter
    def time_unit(self, unit):
        if unit is None:
            unit = self._cfg['units']['defaults']['fluxes']['time']
        if not isinstance(unit, units.UnitBase):
            raise TypeError(
                'The property time_unit must be of type '
                'astropy.units.UnitBase!')
        self._time_unit = unit

    @property
    def unit_str(self):
        """The string representation of the flux unit.
        """
        if self.angle_unit == units.radian:
            angle_unit_sq = units.steradian
        else:
            angle_unit_sq = self.angle_unit**2

        s = (f'({self.energy_unit.to_string()}'
             f' {angle_unit_sq.to_string()}'
             f' {self.length_unit.to_string()}^2'
             f' {self.time_unit.to_string()})^-1')

        return s

    @property
    def unit_latex_str(self):
        """The latex string representation of the flux unit.
        """
        if self.angle_unit == units.radian:
            angle_unit_sq = units.steradian
        else:
            angle_unit_sq = self.angle_unit**2

        s = (f'{self.energy_unit.to_string()}''$^{-1}$ '
             f'{angle_unit_sq.to_string()}''$^{-1}$ '
             f'{self.length_unit.to_string()}''$^{-2}$ '
             f'{self.time_unit.to_string()}''$^{-1}$')

        return s

    def __str__(self):
        """Pretty string representation of this class.
        """
        return f'{self.math_function_str} {self.unit_str}'

    @abc.abstractmethod
    def __call__(
            self,
            ra=None,
            dec=None,
            E=None,
            t=None,
            angle_unit=None,
            energy_unit=None,
            time_unit=None):
        """The call operator to retrieve a flux value for a given celestrial
        position, energy, and observation time.

        Parameters
        ----------
        ra : float | (Ncoord,)-shaped 1D numpy ndarray of float
            The right-ascention coordinate for which to retrieve the flux value.
        dec : float | (Ncoord,)-shaped 1D numpy ndarray of float
            The declination coordinate for which to retrieve the flux value.
        E : float | (Nenergy,)-shaped 1D numpy ndarray of float
            The energy for which to retrieve the flux value.
        t : float | (Ntime,)-shaped 1D numpy ndarray of float
            The observation time for which to retrieve the flux value.
        angle_unit : instance of astropy.units.UnitBase | None
            The unit of the given angles.
            If ``None``, the set angle unit of the flux model is assumed.
        energy_unit : instance of astropy.units.UnitBase | None
            The unit of the given energies.
            If ``None``, the set energy unit of the flux model is assumed.
        time_unit : instance of astropy.units.UnitBase | None
            The unit of the given times.
            If ``None``, the set time unit of the flux model is assumed.

        Returns
        -------
        flux : (Ncoord,Nenergy,Ntime)-shaped ndarray of float
            The flux values are in unit of the set flux model units
            [energy]^{-1} [angle]^{-2} [length]^{-2} [time]^{-1}.
        """
        pass

    def to_internal_flux_unit(self):
        """Calculates the conversion factor to convert the flux unit of this
        flux model instance to the SkyLLH internally used flux unit.

        Returns
        -------
        factor : float
            The conversion factor.
        """
        self_flux_unit = 1 / (
            self.angle_unit**2 *
            self.energy_unit *
            self.length_unit**2 *
            self.time_unit)

        internal_units = self._cfg['units']['internal']
        internal_flux_unit = 1 / (
            internal_units['angle']**2 *
            internal_units['energy'] *
            internal_units['length']**2 *
            internal_units['time'])

        factor = (self_flux_unit).to(internal_flux_unit).value

        return factor


class NullFluxModel(
        FluxModel,
):
    """This class provides a dummy flux model class, which can be used for
    testing purposes, in cases where an actual flux model is not required but
    the framework interface requires one.
    """
    def __init__(
            self,
            *args,
            cfg=None,
            **kwargs,
    ):
        """Creates a new instance of NullFluxModel.

        Parameters
        ----------
        cfg : instance of Config | None
            The instance of Config holding the local configuration. Since this
            flux model does nothing, this argument is optional. If not provided
            the default configuration is used.
        """
        if cfg is None:
            cfg = Config()

        super().__init__(
            *args,
            cfg=cfg,
            **kwargs)

    def math_function_str(self):
        """Since this is a dummy flux model, calling this method will raise a
        NotImplementedError.
        """
        raise NotImplementedError(
            f'The {classname(self)} flux model is a dummy flux model which has '
            'no math function prepresentation!')

    def __call__(self, *args, **kwargs):
        """Since this is a dummy flux model, calling this method will raise a
        NotImplementedError.
        """
        raise NotImplementedError(
            f'The {classname(self)} flux model is a dummy flux model and '
            'cannot be called!')


class FactorizedFluxModel(
        FluxModel,
):
    r"""This class describes a flux model where the spatial, energy, and time
    profiles of the source factorize. That means the flux can be written as:

    .. math::

        \Phi(\alpha,\delta,E,t | \vec{p}_\mathrm{s}) =
            \Phi_0
            \Psi(\alpha,\delta|\vec{p}_\mathrm{s})
            \epsilon(E|\vec{p}_\mathrm{s})
            T(t|\vec{p}_\mathrm{s})

    where, :math:`\Phi_0` is the normalization constant of the flux, and
    :math:`\Psi`, :math:`\epsilon`, and :math:`T` are the spatial, energy, and
    time profiles of the flux given the source parameters
    :math:`\vec{p}_\mathrm{s}`, respectively.
    """
    def __init__(
            self,
            Phi0,
            spatial_profile,
            energy_profile,
            time_profile,
            length_unit=None,
            **kwargs,
    ):
        """Creates a new factorized flux model.

        Parameters
        ----------
        Phi0 : float
            The flux normalization constant.
        spatial_profile : instance of SpatialFluxProfile | None
            The SpatialFluxProfile instance providing the spatial profile
            function of the flux.
            If set to None, an instance of UnitySpatialFluxProfile will be used,
            which represents the constant function 1.
        energy_profile : instance of EnergyFluxProfile | None
            The EnergyFluxProfile instance providing the energy profile
            function of the flux.
            If set to None, an instance of UnityEnergyFluxProfile will be used,
            which represents the constant function 1.
        time_profile : instance of TimeFluxProfile | None
            The TimeFluxProfile instance providing the time profile function
            of the flux.
            If set to None, an instance of UnityTimeFluxProfile will be used,
            which represents the constant function 1.
        length_unit : instance of astropy.units.UnitBase | None
            The used unit for length.
            If set to ``None``, the configured default length unit for fluxes is
            used.
        """
        cfg = kwargs.get('cfg')

        self.Phi0 = Phi0

        if spatial_profile is None:
            spatial_profile = UnitySpatialFluxProfile(
                cfg=cfg)
        self.spatial_profile = spatial_profile

        if energy_profile is None:
            energy_profile = UnityEnergyFluxProfile(
                cfg=cfg)
        self.energy_profile = energy_profile

        if time_profile is None:
            time_profile = UnityTimeFluxProfile(
                cfg=cfg)
        self.time_profile = time_profile

        # The base class will set the default (internally used) flux unit, which
        # will be set automatically to the particular profile.
        super().__init__(
            angle_unit=self._spatial_profile.angle_unit,
            energy_unit=self._energy_profile.energy_unit,
            time_unit=self._time_profile.time_unit,
            length_unit=length_unit,
            **kwargs,
        )

        # Define the parameters which can be set via the `set_params`
        # method.
        self.param_names = ('Phi0',)

    @property
    def Phi0(self):
        """The flux normalization constant.
        The unit of this normalization constant is
        ([angle]^{-2} [energy]^{-1} [length]^{-2} [time]^{-1}).
        """
        return self._Phi0

    @Phi0.setter
    def Phi0(self, v):
        v = float_cast(
            v,
            'The Phi0 property must be castable to type float!')
        self._Phi0 = v

    @property
    def spatial_profile(self):
        """Instance of SpatialFluxProfile describing the spatial profile of the
        flux.
        """
        return self._spatial_profile

    @spatial_profile.setter
    def spatial_profile(self, profile):
        if not isinstance(profile, SpatialFluxProfile):
            raise TypeError(
                'The spatial_profile property must be None, or an '
                'instance of SpatialFluxProfile!')
        self._spatial_profile = profile

    @property
    def energy_profile(self):
        """Instance of EnergyFluxProfile describing the energy profile of the
        flux.
        """
        return self._energy_profile

    @energy_profile.setter
    def energy_profile(self, profile):
        if not isinstance(profile, EnergyFluxProfile):
            raise TypeError(
                'The energy_profile property must be None, or an '
                'instance of EnergyFluxProfile!')
        self._energy_profile = profile

    @property
    def time_profile(self):
        """Instance of TimeFluxProfile describing the time profile of the flux.
        """
        return self._time_profile

    @time_profile.setter
    def time_profile(self, profile):
        if not isinstance(profile, TimeFluxProfile):
            raise TypeError(
                'The time_profile property must be None, or an '
                'instance of TimeFluxProfile!')
        self._time_profile = profile

    @property
    def math_function_str(self):
        """The string showing the mathematical function of the flux.
        """
        s = f'{self._Phi0:.3e}'

        spatial_str = self._spatial_profile.math_function_str
        if spatial_str is not None:
            s += f' * {spatial_str}'
        energy_str = self._energy_profile.math_function_str
        if energy_str is not None:
            s += f' * {energy_str}'

        time_str = self._time_profile.math_function_str
        if time_str is not None:
            s += f' * {time_str}'

        return s

    @property
    def angle_unit(self):
        """The unit of angle used for the flux calculation. The unit is
        taken and set from and to the set spatial flux profile, respectively.
        """
        return self._spatial_profile.angle_unit

    @angle_unit.setter
    def angle_unit(self, unit):
        self._spatial_profile.angle_unit = unit

    @property
    def energy_unit(self):
        """The unit of energy used for the flux calculation. The unit is
        taken and set from and to the set energy flux profile, respectively.
        """
        return self._energy_profile.energy_unit

    @energy_unit.setter
    def energy_unit(self, unit):
        self._energy_profile.energy_unit = unit

    @property
    def time_unit(self):
        """The unit of time used for the flux calculation. The unit is
        taken and set from and to the set time flux profile, respectively.
        """
        return self._time_profile.time_unit

    @time_unit.setter
    def time_unit(self, unit):
        self._time_profile.time_unit = unit

    @property
    def param_names(self):
        """The tuple holding the names of the math function's parameters. This
        is the total set of parameter names for all flux profiles of this
        FactorizedFluxModel instance.
        """
        pnames = list(super(FactorizedFluxModel, type(self)).param_names)
        pnames += self._spatial_profile.param_names
        pnames += self._energy_profile.param_names
        pnames += self._time_profile.param_names

        return tuple(pnames)

    @param_names.setter
    def param_names(self, names):
        super(FactorizedFluxModel, type(self)).param_names.fset(self, names)

    def __call__(
            self,
            ra=None,
            dec=None,
            E=None,
            t=None,
            angle_unit=None,
            energy_unit=None,
            time_unit=None,
    ):
        """Calculates the flux values for the given celestrial positions,
        energies, and observation times.

        Parameters
        ----------
        ra: float | (Ncoord,)-shaped 1d numpy ndarray of float | None
            The right-ascention coordinate for which to retrieve the flux value.
        dec : float | (Ncoord,)-shaped 1d numpy ndarray of float | None
            The declination coordinate for which to retrieve the flux value.
        E : float | (Nenergy,)-shaped 1d numpy ndarray of float | None
            The energy for which to retrieve the flux value.
        t : float | (Ntime,)-shaped 1d numpy ndarray of float | None
            The observation time for which to retrieve the flux value.
        angle_unit : instance of astropy.units.UnitBase | None
            The unit of the given angles.
            If ``None``, the set angle unit of the spatial flux profile is
            assumed.
        energy_unit : instance of astropy.units.UnitBase | None
            The unit of the given energies.
            If ``None``, the set energy unit of the energy flux profile is
            assumed.
        time_unit : instance of astropy.units.UnitBase | None
            The unit of the given times.
            If ``None``, the set time unit of the time flux profile is
            assumed.

        Returns
        -------
        flux : (Ncoord,Nenergy,Ntime)-shaped ndarray of float
            The flux values are in unit
            [energy]^{-1} [angle]^{-2} [length]^{-2} [time]^{-1}.
        """
        if (ra is not None) and (dec is not None):
            spatial_profile_values = self._spatial_profile(
                ra, dec, unit=angle_unit)
        else:
            spatial_profile_values = np.array([1])

        if E is not None:
            energy_profile_values = self._energy_profile(
                E, unit=energy_unit)
        else:
            energy_profile_values = np.array([1])

        if t is not None:
            time_profile_values = self._time_profile(
                t, unit=time_unit)
        else:
            time_profile_values = np.array([1])

        flux = (
            self._Phi0 *
            spatial_profile_values[:, np.newaxis, np.newaxis] *
            energy_profile_values[np.newaxis, :, np.newaxis] *
            time_profile_values[np.newaxis, np.newaxis, :]
        )

        return flux

    def get_param(self, name):
        """Retrieves the value of the given parameter. It returns ``np.nan`` if
        the parameter does not exist.

        Parameters
        ----------
        name : str
            The name of the parameter.

        Returns
        -------
        value : float | np.nan
            The value of the parameter.
        """
        for obj in (
                super(),
                self._spatial_profile,
                self._energy_profile,
                self._time_profile):
            value = obj.get_param(name=name)
            if not np.isnan(value):
                return value

        return np.nan

    def set_params(self, pdict):
        """Sets the parameters of the flux model. For this factorized flux model
        it means that it sets the parameters of the spatial, energy, and time
        profiles.

        Parameters
        ----------
        pdict : dict
            The flux parameter dictionary.

        Returns
        -------
        updated : bool
            Flag if parameter values were actually updated.
        """
        updated = False

        updated |= super().set_params(pdict)

        updated |= self._spatial_profile.set_params(pdict)
        updated |= self._energy_profile.set_params(pdict)
        updated |= self._time_profile.set_params(pdict)

        return updated


class PointlikeFFM(
        FactorizedFluxModel,
        IsPointlike,
):
    """This class describes a factorized flux model (FFM), where the spatial
    profile is modeled as a point. This class provides the base class for a flux
    model of a point-like source.
    """
    def __init__(
            self,
            Phi0,
            energy_profile,
            time_profile,
            ra=None,
            dec=None,
            angle_unit=None,
            length_unit=None,
            **kwargs,
    ):
        """Creates a new factorized flux model for a point-like source.

        Parameters
        ----------
        Phi0 : float
            The flux normalization constant in unit of flux.
        energy_profile : instance of EnergyFluxProfile | None
            The EnergyFluxProfile instance providing the energy profile
            function of the flux.
            If set to None, an instance of UnityEnergyFluxProfile will be used,
            which represents the constant function 1.
        time_profile : instance of TimeFluxProfile | None
            The TimeFluxProfile instance providing the time profile function
            of the flux.
            If set to None, an instance of UnityTimeFluxProfile will be used,
            which represents the constant function 1.
        ra : float | None
            The right-ascention of the point.
        dec : float | None
            The declination of the point.
        angle_unit : instance of astropy.units.UnitBase | None
            The unit for angles used for the flux unit.
            If set to ``None``, the configured internal angle unit is used.
        length_unit : instance of astropy.units.UnitBase | None
            The unit for length used for the flux unit.
            If set to ``None``, the configured internal length unit is used.
        """
        spatial_profile = PointSpatialFluxProfile(
            cfg=kwargs.get('cfg'),
            ra=ra,
            dec=dec,
            angle_unit=angle_unit)

        super().__init__(
            Phi0=Phi0,
            spatial_profile=spatial_profile,
            energy_profile=energy_profile,
            time_profile=time_profile,
            length_unit=length_unit,
            ra_func_instance=spatial_profile,
            get_ra_func=type(spatial_profile).ra.fget,
            set_ra_func=type(spatial_profile).ra.fset,
            dec_func_instance=spatial_profile,
            get_dec_func=type(spatial_profile).dec.fget,
            set_dec_func=type(spatial_profile).dec.fset,
            **kwargs)

    @property
    def unit_str(self):
        """The string representation of the flux unit.
        """
        # Note:
        #    For a point-like differential flux, there is no solid-angle
        #    element.
        s = (f'({self.energy_unit.to_string()}'
             f' {self.length_unit.to_string()}^2'
             f' {self.time_unit.to_string()})^-1')

        return s

    @property
    def unit_latex_str(self):
        """The latex string representation of the flux unit.
        """
        # Note:
        #    For a point-like differential flux, there is no solid-angle
        #    element.
        s = (f'{self.energy_unit.to_string()}''$^{-1}$ '
             f'{self.length_unit.to_string()}''$^{-2}$ '
             f'{self.time_unit.to_string()}''$^{-1}$')

        return s


class SteadyPointlikeFFM(
        PointlikeFFM,
):
    """This class describes a factorized flux model (FFM), where the spatial
    profile is modeled as a point and the time profile as constant 1. It is
    derived from the ``PointlikeFFM`` class.
    """
    def __init__(
            self,
            Phi0,
            energy_profile,
            ra=None,
            dec=None,
            angle_unit=None,
            length_unit=None,
            time_unit=None,
            **kwargs,
    ):
        """Creates a new factorized flux model for a point-like source with no
        time dependance.

        Parameters
        ----------
        Phi0 : float
            The flux normalization constant.
        energy_profile : instance of EnergyFluxProfile | None
            The EnergyFluxProfile instance providing the energy profile
            function of the flux.
            If set to None, an instance of UnityEnergyFluxProfile will be used,
            which represents the constant function 1.
        ra : float | None
            The right-ascention of the point.
        dec : float | None
            The declination of the point.
        angle_unit : instance of astropy.units.UnitBase | None
            The unit for angles used for the flux unit.
            If set to ``None``, the configured default angle unit for fluxes
            is used.
        length_unit : instance of astropy.units.UnitBase | None
            The unit for length used for the flux unit.
            If set to ``None``, the configured default length unit for fluxes
            is used.
        time_unit : instance of astropy.units.UnitBase | None
            The used unit for time.
            If set to ``None``, the configured default time unit for fluxes
            is used.
        """
        time_profile = UnityTimeFluxProfile(
            cfg=kwargs.get('cfg'),
            time_unit=time_unit)

        super().__init__(
            Phi0=Phi0,
            ra=ra,
            dec=dec,
            energy_profile=energy_profile,
            time_profile=time_profile,
            angle_unit=angle_unit,
            length_unit=length_unit,
            **kwargs)



# -*- coding: utf-8 -*-

import numpy as np
import pickle

from skyllh.core.binning import (
    get_bincenters_from_binedges,
)
from skyllh.core.flux_model import (
    PowerLawEnergyFluxProfile,
    SteadyPointlikeFFM,
)


def get_dOmega(dec_min, dec_max):
    """Calculates the solid angle given two declination angles.

    Parameters
    ----------
    dec_min : float | array of float
        The smaller declination angle.
    dec_max : float | array of float
        The larger declination angle.

    Returns
    -------
    solidangle : float | array of float
        The solid angle corresponding to the two given declination angles.
    """
    return 2*np.pi*(np.sin(dec_max) - np.sin(dec_min))


def southpole_zen2dec(zen):
    """Converts zenith angles at the South Pole to declination angles.

    Parameters
    ----------
    zen : (n,)-shaped 1d numpy ndarray
        The numpy ndarray holding the zenith angle values in radians.

    Returns
    -------
    dec : (n,)-shaped 1d numpy ndarray
        The numpy ndarray holding the declination angle values in radians.
    """
    dec = zen - np.pi/2
    return dec


def get_flux_atmo_decnu_log10enu(flux_pathfilename, log10_enu_max=9):
    """Constructs the atmospheric flux map function
    f_atmo(log10(E_nu/GeV),dec_nu) in unit 1/(GeV cm^2 sr s).

    Parameters
    ----------
    flux_pathfilename : str
        The pathfilename of the file containing the MCEq fluxes.
    log10_enu_max : float
        The log10(E/GeV) value of the maximum neutrino energy to be considered.

    Returns
    -------
    flux_atmo : (n_dec, n_e_grid)-shaped 2D numpy ndarray
        The numpy ndarray holding the the atmospheric neutrino flux function in
        unit 1/(GeV cm^2 sr s).
    decnu_binedges : (n_decnu+1,)-shaped 1D numpy ndarray
        The numpy ndarray holding the dec_nu bin edges.
    log10_enu_binedges : (n_enu+1,)-shaped 1D numpy ndarray
        The numpy ndarray holding the neutrino energy bin edges in log10.
    """
    with open(flux_pathfilename, 'rb') as f:
        ((e_grid, zenith_angle_binedges), flux_def) = pickle.load(f)
    zenith_angle_binedges = np.deg2rad(zenith_angle_binedges)

    # Select energy bins below 10**log10_true_e_max GeV.
    m_e_grid = e_grid <= 10**log10_enu_max
    e_grid = e_grid[m_e_grid]

    decnu_binedges = southpole_zen2dec(zenith_angle_binedges)
    decnu_angles = get_bincenters_from_binedges(decnu_binedges)

    # Calculate the neutrino energy bin edges in log10.
    log10_enu_binedges = np.empty((len(e_grid)+1),)
    d_log10_enu = np.diff(np.log10(e_grid))[0]
    log10_enu_binedges[:-1] = np.log10(e_grid) - d_log10_enu/2
    log10_enu_binedges[-1] = log10_enu_binedges[-2] + d_log10_enu

    n_decnu = len(decnu_angles)
    n_enu = len(e_grid)

    # Calculate f_atmo(E_nu,dec_nu).
    f_atmo = np.zeros((n_decnu, n_enu))
    zero_zen_idx = np.digitize(0, zenith_angle_binedges) - 1
    for (decnu_idx, decnu) in enumerate(decnu_angles):
        if decnu < 0:
            fl = flux_def['numu_total'][:, decnu_idx][m_e_grid]
        else:
            # For up-going we use the flux calculation from the streight
            # downgoing.
            fl = flux_def['numu_total'][:, zero_zen_idx][m_e_grid]
        f_atmo[decnu_idx] = fl

    return (f_atmo, decnu_binedges, log10_enu_binedges)


def get_flux_astro_decnu_log10enu(decnu_binedges, log10_enu_binedges):
    """Constructs the astrophysical neutrino flux function
    f_astro(log10(E_nu/GeV),dec_nu) in unit 1/(GeV cm^2 sr s).

    It uses the best fit from the IceCube publication [1].

    Parameters
    ----------
    decnu_binedges : (n_decnu+1,)-shaped 1D numpy ndarray
        The numpy ndarray holding the dec_nu bin edges.
    log10_enu_binedges : (n_enu+1,)-shaped 1D numpy ndarray
        The numpy ndarray holding the log10 values of the neutrino energy bin
        edges in GeV.

    Returns
    -------
    f_astro : (n_decnu, n_log10enu)-shaped 2D numpy ndarray
        The numpy ndarray holding the astrophysical flux values in unit
        1/(GeV cm^2 sr s).

    References
    ----------
    [1] https://arxiv.org/pdf/2111.10299.pdf
    """
    fluxmodel = SteadyPointlikeFFM(
        Phi0=1.44e-18,
        energy_profile=PowerLawEnergyFluxProfile(
            E0=100e3,
            gamma=2.37))

    n_decnu = len(decnu_binedges) - 1

    enu_binedges = np.power(10, log10_enu_binedges)
    enu_bincenters = get_bincenters_from_binedges(enu_binedges)

    fl = fluxmodel(E=enu_bincenters).squeeze()
    f_astro = np.tile(fl, (n_decnu, 1))

    return f_astro


def convert_flux_bkg_to_pdf_bkg(f_bkg, decnu_binedges, log10_enu_binedges):
    """Converts the given background flux function f_bkg into a background flux
    PDF in unit 1/(log10(E/GeV) rad).

    Parameters
    ----------
    f_bkg : (n_decnu, n_enu)-shaped 2D numpy ndarray
        The numpy ndarray holding the background flux values in unit
        1/(GeV cm^2 s sr).
    decnu_binedges : (n_decnu+1,)-shaped 1D numpy ndarray
        The numpy ndarray holding the dec_nu bin edges in radians.
    log10_enu_binedges : (n_enu+1,)-shaped 1D numpy ndarray
        The numpy ndarray holding the log10 values of the neutrino energy bin
        edges in GeV.

    Returns
    -------
    p_bkg : (n_decnu, n_enu)-shaped 2D numpy ndarray
        The numpy ndarray holding the background flux pdf values.
    """
    d_decnu = np.diff(decnu_binedges)
    d_log10_enu = np.diff(log10_enu_binedges)

    bin_area = d_decnu[:, np.newaxis] * d_log10_enu[np.newaxis, :]
    p_bkg = f_bkg / np.sum(f_bkg*bin_area)

    # Cross-check the normalization of the PDF.
    if not np.isclose(np.sum(p_bkg*bin_area), 1):
        raise ValueError(
            'The background PDF is not normalized! The integral is '
            f'{np.sum(p_bkg*bin_area)}!')

    return p_bkg


def get_pd_atmo_decnu_Enu(flux_pathfilename, log10_true_e_max=9):
    """Constructs the atmospheric neutrino PDF p_atmo(E_nu,dec_nu) in unit
    1/(GeV rad).

    Parameters
    ----------
    flux_pathfilename : str
        The pathfilename of the file containing the MCEq flux.
    log10_true_e_max : float
        The log10(E/GeV) value of the maximum true energy to be considered.

    Returns
    -------
    pd_atmo : (n_dec, n_e_grid)-shaped 2D numpy ndarray
        The numpy ndarray holding the the atmospheric neutrino PDF in unit
        1/(GeV rad).
    decnu_binedges : (n_decnu+1,)-shaped 1D numpy ndarray
        The numpy ndarray holding the dec_nu bin edges.
    log10_e_grid_edges : (n_e_grid+1,)-shaped 1D numpy ndarray
        The numpy ndarray holding the energy bin edges in log10.
    """
    with open(flux_pathfilename, 'rb') as f:
        ((e_grid, zenith_angle_binedges), flux_def) = pickle.load(f)

    # Select energy bins below 10**log10_true_e_max GeV.
    m_e_grid = e_grid <= 10**log10_true_e_max
    e_grid = e_grid[m_e_grid]

    zenith_angles = 0.5*(zenith_angle_binedges[:-1] + zenith_angle_binedges[1:])
    decnu_angles = np.deg2rad(zenith_angles) - np.pi/2

    decnu_binedges = np.deg2rad(zenith_angle_binedges) - np.pi/2
    d_decnu = np.diff(decnu_binedges)

    # Calculate the e_grid bin edges in log10.
    log10_e_grid_edges = np.empty((len(e_grid)+1),)
    d_log10_e_grid = np.diff(np.log10(e_grid))[0]
    log10_e_grid_edges[:-1] = np.log10(e_grid) - d_log10_e_grid/2
    log10_e_grid_edges[-1] = log10_e_grid_edges[-2] + d_log10_e_grid

    n_decnu = len(decnu_angles)
    n_e_grid = len(e_grid)

    # Calculate p_atmo(E_nu,dec_nu).
    pd_atmo = np.zeros((n_decnu, n_e_grid))
    for (decnu_idx, decnu) in enumerate(decnu_angles):
        if decnu < 0:
            fl = flux_def['numu_total'][:, decnu_idx][m_e_grid]
        else:
            # For up-going we use the flux calculation from the streight
            # downgoing.
            fl = flux_def['numu_total'][:, 0][m_e_grid]
        pd_atmo[decnu_idx] = fl
    # Normalize the PDF.
    bin_area = d_decnu[:, np.newaxis] * np.diff(log10_e_grid_edges)[np.newaxis, :]
    pd_atmo /= np.sum(pd_atmo*bin_area)

    # Cross-check the normalization of the PDF.
    if not np.isclose(np.sum(pd_atmo*bin_area), 1):
        raise ValueError(
            'The atmospheric true energy PDF is not normalized! The integral '
            f'is {np.sum(pd_atmo*bin_area)}!')

    return (pd_atmo, decnu_binedges, log10_e_grid_edges)


def get_pd_atmo_E_nu_sin_dec_nu(flux_pathfilename):
    """Constructs the atmospheric energy PDF p_atmo(E_nu|sin(dec_nu)) in
    unit 1/GeV.

    Parameters
    ----------
    flux_pathfilename : str
        The pathfilename of the file containing the MCEq flux.

    Returns
    -------
    pd_atmo : (n_sin_dec, n_e_grid)-shaped 2D numpy ndarray
        The numpy ndarray holding the the atmospheric energy PDF in unit 1/GeV.
    sin_dec_binedges : numpy ndarray
        The (n_sin_dec+1,)-shaped 1D numpy ndarray holding the sin(dec) bin
        edges.
    log10_e_grid_edges : numpy ndarray
        The (n_e_grid+1,)-shaped 1D numpy ndarray holding the energy bin edges
        in log10.
    """
    with open(flux_pathfilename, 'rb') as f:
        ((e_grid, zenith_angle_binedges), flux_def) = pickle.load(f)

    # Select energy bins below 10**9 GeV.
    m_e_grid = e_grid <= 10**9
    e_grid = e_grid[m_e_grid]

    zenith_angles = 0.5*(zenith_angle_binedges[:-1] + zenith_angle_binedges[1:])

    # Calculate the e_grid bin edges in log10.
    log10_e_grid_edges = np.empty((len(e_grid)+1),)
    d_log10_e_grid = np.diff(np.log10(e_grid))[0]
    log10_e_grid_edges[:-1] = np.log10(e_grid) - d_log10_e_grid/2
    log10_e_grid_edges[-1] = log10_e_grid_edges[-2] + d_log10_e_grid

    # Calculate the energy bin widths of the energy grid.
    dE = np.diff(10**log10_e_grid_edges)

    # Convert zenith angles into sin(declination) angles.
    sin_dec_binedges = np.sin(np.deg2rad(zenith_angle_binedges) - np.pi/2)
    sin_dec_angles = np.sin(np.deg2rad(zenith_angles) - np.pi/2)

    n_e_grid = len(e_grid)
    n_sin_dec = len(sin_dec_angles)

    # Calculate p_atmo(E_nu|sin(dec_nu)).
    pd_atmo = np.zeros((n_sin_dec, n_e_grid))
    for (sin_dec_idx, sin_dec) in enumerate(sin_dec_angles):
        if sin_dec < 0:
            fl = flux_def['numu_total'][:, sin_dec_idx][m_e_grid]
        else:
            # For up-going we use the flux calculation from the streight
            # downgoing.
            fl = flux_def['numu_total'][:, 0][m_e_grid]
        pd_atmo[sin_dec_idx] = fl/np.sum(fl*dE)

    # Cross-check the normalization of the PDF.
    if not np.all(np.isclose(np.sum(pd_atmo*dE[np.newaxis, :], axis=1), 1)):
        raise ValueError(
            'The atmospheric true energy PDF is not normalized!')

    return (pd_atmo, sin_dec_binedges, log10_e_grid_edges)


def get_pd_astro_E_nu_sin_dec_nu(sin_dec_binedges, log10_e_grid_edges):
    """Constructs the astrophysical energy PDF p_astro(E_nu|sin(dec_nu)) in
    unit 1/GeV.
    It uses the best fit from the IceCube publication [1].

    Parameters
    ----------
    sin_dec_binedges : (n_sin_dec+1,)-shaped 1D numpy ndarray
        The numpy ndarray holding the sin(dec) bin edges.
    log10_e_grid_edges : (n_e_grid+1,)-shaped 1D numpy ndarray
        The numpy ndarray holding the log10 values of the energy bin edges in
        GeV of the energy grid.

    Returns
    -------
    pd_astro : (n_sin_dec, n_e_grid)-shaped 2D numpy ndarray
        The numpy ndarray holding the energy probability density values
        p(E_nu|sin_dec_nu) in unit 1/GeV.

    References
    ----------
    [1] https://arxiv.org/pdf/2111.10299.pdf
    """
    fluxmodel = SteadyPointlikeFFM(
        Phi0=1.44e-18,
        energy_profile=PowerLawEnergyFluxProfile(
            E0=100e3,
            gamma=2.37))

    n_sin_dec = len(sin_dec_binedges) - 1

    e_grid_edges = 10**log10_e_grid_edges
    e_grid_bc = 0.5*(e_grid_edges[:-1] + e_grid_edges[1:])

    dE = np.diff(e_grid_edges)

    fl = fluxmodel(E=e_grid_bc).squeeze()
    pd = fl / np.sum(fl*dE)
    pd_astro = np.tile(pd, (n_sin_dec, 1))

    # Cross-check the normalization of the PDF.
    if not np.all(np.isclose(np.sum(pd_astro*dE[np.newaxis, :], axis=1), 1)):
        raise ValueError(
            'The astrophysical energy PDF is not normalized!')

    return pd_astro


def get_pd_bkg_E_nu_sin_dec_nu(pd_atmo, pd_astro, log10_e_grid_edges):
    """Constructs the total background flux probability density
    p_bkg(E_nu|sin(dec_nu)) in unit 1/GeV.

    Parameters
    ----------
    pd_atmo : (n_sin_dec, n_e_grid)-shaped 2D numpy ndarray
        The numpy ndarray holding the probability density values
        p(E_nu|sin(dec_nu)) in 1/GeV of the atmospheric flux.
    pd_astro : (n_sin_dec, n_e_grid)-shaped 2D numpy ndarray
        The numpy ndarray holding the probability density values
        p(E_nu|sin(dec_nu)) in 1/GeV of the astrophysical flux.
    log10_e_grid_edges : (n_e_grid+1,)-shaped numpy ndarray
        The numpy ndarray holding the log10 values of the energy grid bin edges
        in GeV.

    Returns
    -------
    pd_bkg : (n_sin_dec, n_e_grid)-shaped 2D numpy ndarray
        The numpy ndarray holding total background probability density values
        p_bkg(E_nu|sin(dec_nu)) in unit 1/GeV.
    """
    pd_bkg = pd_atmo + pd_astro

    dE = np.diff(10**log10_e_grid_edges)

    s = np.sum(pd_bkg*dE[np.newaxis, :], axis=1, keepdims=True)
    pd_bkg /= s

    if not np.all(np.isclose(np.sum(pd_bkg*dE[np.newaxis, :], axis=1), 1)):
        raise ValueError(
            'The background energy PDF is not normalized!')

    return pd_bkg



# -*- coding: utf-8 -*-
# Author: Dr. Martin Wolf <mail@martin-wolf.org>

import matplotlib as mpl
import numpy as np

from matplotlib import (
    pyplot as plt,
)
from mpl_toolkits.axes_grid1.axes_divider import (
    make_axes_locatable,
)


def plot_ns_fit_vs_mean_ns_inj(  # noqa: C901
        trials,
        mean_n_sig_key='mean_n_sig',
        ns_fit_key='ns',
        rethist=False,
        title='',
        figsize=None,
        line_color=None,
        axis_fontsize=16,
        title_fontsize=16,
        tick_fontsize=16,
        xlabel=None,
        ylabel=None,
        ylim=None,
        ratio_ylim=None,
):
    r"""Creates a 2D histogram plot showing the fit number of signal events vs.
    the mean number of injected signal events.

    Parameters
    ----------
    trials : numpy record array
        The record array holding the results of the trials.
    mean_n_sig_key : str
        The name of the key for the mean number of injected signal events in the
        given trials record array.
        Default is ``'mean_n_sig'``.
    ns_fit_key : str
        The name of the key for the fitted number of signal events in the
        given trials record array.
        Default is ``'ns'``.
    rethist : bool
        If set to ``True``, the histogram data along with the histogram bin
        edges will be return as well.
        Default is ``False``.
    title : str
        The title of the plot.
    figsize : tuple | None
        The two-element tuple (width,height) specifying the size of the figure.
        If set to None, the default size (12,10) will be used.
    line_color : str | None
        The color of the lines.
        The default is '#E37222'.
    axis_fontsize : float
        The font size of the axis labels.
    title_fontsize : float
        The font size of the plot title.
    tick_fontsize : float
        The font size of the tick labels.
    xlabel : str | None
        The label of the x-axis in math syntax.
        Default is ``r'<n>_{\mathrm{sig,inj}}}'``.
    ylabel : str | None
        The label of if y-axis in math syntax.
        Default is ``r'n_\mathrm{sig,fit}'``.
    ylim : tuple | None
        The (low,high)-two-element tuple specifying the y-axis limits of the
        main plot.
    ratio_ylim : tuple | None
        The (low,high)-two-element tuple specifying the y-axis limits of the
        ratio plot in percentage.
        If set to None, the default (-100,100) will be used.

    Returns
    -------
    fig : MPL Figure instance
        The created matplotlib Figure instance.
    hist : 2d ndarray
        The histogram bin content.
        This will only be returned, when the ``rethist`` argument was set to
        ``True``.
    xedges : 1d ndarray
        The histogram x-axis bin edges.
        This will only be returned, when the ``rethist`` argument was set to
        ``True``.
    yedges : 1d ndarray
        The histogram y-axis bin edges.
        This will only be returned, when the ``rethist`` argument was set to
        ``True``.
    """
    if figsize is None:
        figsize = (12, 10)
    if line_color is None:
        line_color = '#E37222'
    if xlabel is None:
        xlabel = r'<n>_{\mathrm{sig,inj}}'
    if ylabel is None:
        ylabel = r'n_\mathrm{sig,fit}'
    if ratio_ylim is None:
        ratio_ylim = (-100, 100)

    # Create the x-axis binning.
    mean_n_sig_min = np.min(trials[mean_n_sig_key])
    mean_n_sig_max = np.max(trials[mean_n_sig_key])
    mean_n_sig_step = np.diff(np.sort(np.unique(trials[mean_n_sig_key])))[0]
    x_bins = np.arange(
        mean_n_sig_min-mean_n_sig_step/2,
        mean_n_sig_max+mean_n_sig_step/2+1,
        mean_n_sig_step)

    # Create the y-axis binning.
    ns_fit_min = np.min(trials[ns_fit_key])
    ns_fit_max = np.max(trials[ns_fit_key])
    dy = np.min((mean_n_sig_step, np.abs(ns_fit_max-ns_fit_min)/100))
    y_bins = np.arange(np.floor(ns_fit_min), ns_fit_max+dy, dy)

    # Calculate the weight of each trial for the histogram so that the trials
    # are normalized separately for each mean number of injected signal events.
    # Also calculate the median and upper and lower 68% quantile of ns_fit.
    hist_weights = np.ones_like(trials[mean_n_sig_key])
    (mean_n_sig, n_trials) = np.unique(
        trials[mean_n_sig_key], return_counts=True)
    ns_fit_median = np.empty_like(mean_n_sig, dtype=np.float64)
    ns_fit_uq = np.empty_like(mean_n_sig, dtype=np.float64)
    ns_fit_lq = np.empty_like(mean_n_sig, dtype=np.float64)
    for (idx, (mean_n_sig_, n_trials_)) in enumerate(zip(mean_n_sig, n_trials)):
        m = trials[mean_n_sig_key] == mean_n_sig_
        hist_weights[m] /= n_trials_
        ns_fit_median[idx] = np.median(trials[m][ns_fit_key])
        ns_fit_uq[idx] = np.percentile(trials[m][ns_fit_key], 84.1)
        ns_fit_lq[idx] = np.percentile(trials[m][ns_fit_key], 15.9)

    # Create two Axes objects, one for the histogram and one for the ratio.
    (fig, ax) = plt.subplots(
        2, 1,
        gridspec_kw={'height_ratios': [3, 1]},
        sharex=True,
        figsize=figsize)

    # Add an axes above the main axes for the colorbar.
    ax_divider = make_axes_locatable(ax[0])
    cax = ax_divider.append_axes("top", size="7%", pad="2%")

    # Create and plot the 2D histogram.
    (hist, xedges, yedges, image) = ax[0].hist2d(
        trials[mean_n_sig_key], trials[ns_fit_key],
        bins=[x_bins, y_bins],
        weights=hist_weights,
        norm=mpl.colors.LogNorm(),
        cmap=plt.get_cmap('GnBu'))

    ax[0].set_ylabel('$'+ylabel+'$', fontsize=axis_fontsize)

    # Add the diagonal expectation line.
    ax[0].plot(
        x_bins, x_bins,
        color='black',
        alpha=0.4,
        linestyle='-',
        linewidth=2)

    # Plot the lower quantile.
    ax[0].plot(
        mean_n_sig, ns_fit_lq,
        color=line_color,
        linestyle='-.',
        linewidth=2)

    # Plot the median fitted ns.
    ax[0].plot(
        mean_n_sig, ns_fit_median,
        color=line_color,
        linestyle='-',
        linewidth=2,
        label=r'median')

    # Plot the upper quantile.
    ax[0].plot(
        mean_n_sig, ns_fit_uq,
        color=line_color,
        linestyle='-.',
        linewidth=2,
        label=r'$1\sigma$')

    ax[0].legend()

    if ylim is not None:
        ax[0].set_ylim(ylim)

    # Create the color bar.
    cb = fig.colorbar(image, cax=cax, orientation='horizontal')
    cb.ax.xaxis.set_ticks_position('top')
    cb.ax.text(
        0.5, 1, title, horizontalalignment='center', verticalalignment='bottom',
        transform=fig.transFigure, fontsize=title_fontsize)

    # Plot the ratio.
    ax[1].hlines(0, x_bins[0], x_bins[-1])
    m = mean_n_sig != 0
    ax[1].plot(
        mean_n_sig[m], (ns_fit_median[m]-mean_n_sig[m])/mean_n_sig[m]*100,
        color=line_color, linewidth=2, label=r'median')
    ax[1].fill_between(
        mean_n_sig[m],
        (ns_fit_uq[m]-mean_n_sig[m])/mean_n_sig[m]*100,
        (ns_fit_lq[m]-mean_n_sig[m])/mean_n_sig[m]*100,
        alpha=0.2, color='gray', label=r'$1\sigma$')

    ax[1].legend()

    ax[1].set_xlabel('$'+xlabel+'$', fontsize=axis_fontsize)
    ratio_ylabel = r'$\frac{%s - %s}{%s}$' % (ylabel, xlabel, xlabel)+' [%]'
    ax[1].set_ylabel(ratio_ylabel, fontsize=axis_fontsize)
    ax[1].set_xlim(x_bins[0], x_bins[-1])
    ax[1].set_ylim(ratio_ylim)

    # Set the font size of the tick labels.
    for tick in ax[0].yaxis.get_major_ticks():
        tick.label.set_fontsize(tick_fontsize)
    for tick in ax[1].xaxis.get_major_ticks():
        tick.label.set_fontsize(tick_fontsize)
    for tick in ax[1].yaxis.get_major_ticks():
        tick.label.set_fontsize(tick_fontsize)

    plt.tight_layout()

    if rethist:
        return (fig, hist, xedges, yedges)

    return fig


def plot_gamma_fit_vs_mean_ns_inj(  # noqa: C901
        trials,
        gamma_inj=2,
        mean_n_sig_key='mean_n_sig',
        gamma_fit_key='gamma',
        rethist=False,
        title='',
        figsize=None,
        line_color=None,
        axis_fontsize=16,
        title_fontsize=16,
        tick_fontsize=16,
        xlabel=None,
        ylabel=None,
        ratio_ylim=None,
):
    r"""Creates a 2D histogram plot showing the fit spectral index gamma vs.
    the mean number of injected signal events.

    Parameters
    ----------
    trials : numpy record array
        The record array holding the results of the trials.
    gamma_inj : float
        The spectral index with which signal events got injected into tha trial
        data set.
    mean_n_sig_key : str
        The name of the key for the mean number of injected signal events in the
        given trials record array.
        Default is ``'mean_n_sig'``.
    gamma_fit_key : str
        The name of the key for the fitted spectral index in the given trials
        record array.
        Default is ``'gamma'``.
    rethist : bool
        If set to ``True``, the histogram data along with the histogram bin
        edges will be return as well.
        Default is ``False``.
    title : str
        The title of the plot.
    figsize : tuple | None
        The two-element tuple (width,height) specifying the size of the figure.
        If set to None, the default size (12,10) will be used.
    line_color : str | None
        The color of the lines.
        The default is '#E37222'.
    axis_fontsize : float
        The font size of the axis labels.
    title_fontsize : float
        The font size of the plot title.
    tick_fontsize : float
        The font size of the tick labels.
    xlabel : str | None
        The label of the x-axis in math syntax.
        Default is ``r'<n>_{\mathrm{sig,inj}}}'``.
    ylabel : str | None
        The label of if y-axis in math syntax.
        Default is ``r'\gamma_\mathrm{fit}'``.
    ratio_ylim : tuple | None
        The (low,high)-two-element tuple specifying the y-axis limits of the
        ratio plot in percentage.
        If set to None, the default (-100,100) will be used.

    Returns
    -------
    fig : MPL Figure instance
        The created matplotlib Figure instance.
    hist : 2d ndarray
        The histogram bin content.
        This will only be returned, when the ``rethist`` argument was set to
        ``True``.
    xedges : 1d ndarray
        The histogram x-axis bin edges.
        This will only be returned, when the ``rethist`` argument was set to
        ``True``.
    yedges : 1d ndarray
        The histogram y-axis bin edges.
        This will only be returned, when the ``rethist`` argument was set to
        ``True``.
    """
    if figsize is None:
        figsize = (12, 10)
    if line_color is None:
        line_color = '#E37222'
    if xlabel is None:
        xlabel = r'<n>_{\mathrm{sig,inj}}'
    if ylabel is None:
        ylabel = r'\gamma_\mathrm{fit}'
    if ratio_ylim is None:
        ratio_ylim = (-100, 100)

    # Create the x-axis binning.
    mean_n_sig_min = np.min(trials[mean_n_sig_key])
    mean_n_sig_max = np.max(trials[mean_n_sig_key])
    mean_n_sig_step = np.diff(np.sort(np.unique(trials[mean_n_sig_key])))[0]
    x_bins = np.arange(
        mean_n_sig_min-mean_n_sig_step/2,
        mean_n_sig_max+mean_n_sig_step/2+1,
        mean_n_sig_step)

    # Create the y-axis binning.
    gamma_fit_min = np.min(trials[gamma_fit_key])
    gamma_fit_max = np.max(trials[gamma_fit_key])
    dy = np.min((mean_n_sig_step, np.abs(gamma_fit_max-gamma_fit_min)/100))
    y_bins = np.arange(np.floor(gamma_fit_min), gamma_fit_max+dy, dy)

    # Calculate the weight of each trial for the histogram so that the trials
    # are normalized separately for each mean number of injected signal events.
    # Also calculate the median and upper and lower 68% quantile of gamma_fit.
    hist_weights = np.ones_like(trials[mean_n_sig_key])
    (mean_n_sig, n_trials) = np.unique(
        trials[mean_n_sig_key], return_counts=True)
    gamma_fit_median = np.empty_like(mean_n_sig, dtype=np.float64)
    gamma_fit_uq = np.empty_like(mean_n_sig, dtype=np.float64)
    gamma_fit_lq = np.empty_like(mean_n_sig, dtype=np.float64)
    for (idx, (mean_n_sig_, n_trials_)) in enumerate(zip(mean_n_sig, n_trials)):
        m = trials[mean_n_sig_key] == mean_n_sig_
        hist_weights[m] /= n_trials_
        gamma_fit_median[idx] = np.median(trials[m][gamma_fit_key])
        gamma_fit_uq[idx] = np.percentile(trials[m][gamma_fit_key], 84.1)
        gamma_fit_lq[idx] = np.percentile(trials[m][gamma_fit_key], 15.9)

    # Create two Axes objects, one for the histogram and one for the ratio.
    (fig, ax) = plt.subplots(
        2, 1,
        gridspec_kw={'height_ratios': [3, 1]},
        sharex=True,
        figsize=figsize)

    # Add an axes above the main axes for the colorbar.
    ax_divider = make_axes_locatable(ax[0])
    cax = ax_divider.append_axes('top', size='7%', pad='2%')

    # Create and plot the 2D histogram.
    (hist, xedges, yedges, image) = ax[0].hist2d(
        trials[mean_n_sig_key], trials[gamma_fit_key],
        bins=[x_bins, y_bins],
        weights=hist_weights,
        norm=mpl.colors.LogNorm(),
        cmap=plt.get_cmap('GnBu'))

    ax[0].set_ylabel('$'+ylabel+'$', fontsize=axis_fontsize)

    # Add the horizontal expectation line.
    ax[0].hlines(
        gamma_inj, x_bins[0], x_bins[-1],
        color='black',
        alpha=0.4,
        linestyle='-',
        linewidth=2)

    # Plot the upper quantile curve.
    ax[0].plot(
        mean_n_sig, gamma_fit_uq,
        color=line_color,
        linestyle='-.',
        linewidth=2)

    # Plot the median fitted gamma.
    ax[0].plot(
        mean_n_sig, gamma_fit_median,
        color=line_color,
        linestyle='-',
        linewidth=2)

    # Plot the lower quantile curve.
    ax[0].plot(
        mean_n_sig, gamma_fit_lq,
        color=line_color,
        linestyle='-.',
        linewidth=2)

    # Create the color bar.
    cb = fig.colorbar(image, cax=cax, orientation='horizontal')
    cb.ax.xaxis.set_ticks_position('top')
    cb.ax.text(
        0.5, 1, title,
        horizontalalignment='center',
        verticalalignment='bottom',
        transform=fig.transFigure,
        fontsize=title_fontsize)

    # Plot the ratio.
    ax[1].hlines(0, x_bins[0], x_bins[-1])
    m = mean_n_sig != 0
    ax[1].plot(
        mean_n_sig[m], (gamma_fit_median[m]-gamma_inj)/gamma_inj*100,
        linewidth=2)
    ax[1].fill_between(
        mean_n_sig[m],
        (gamma_fit_uq[m]-gamma_inj)/gamma_inj*100,
        (gamma_fit_lq[m]-gamma_inj)/gamma_inj*100,
        alpha=0.2,
        color='gray',
        label=r'$1\sigma$')

    ax[1].legend()

    ax[1].set_xlabel('$'+xlabel+'$', fontsize=axis_fontsize)
    gamma_inj_label = r'\gamma_{\mathrm{inj}}'
    ratio_ylabel = r'$\frac{<%s> - %s}{%s}$' % (
        ylabel, gamma_inj_label, gamma_inj_label)+' [%]'
    ax[1].set_ylabel(ratio_ylabel, fontsize=axis_fontsize)
    ax[1].set_xlim(x_bins[0], x_bins[-1])
    ax[1].set_ylim(ratio_ylim)

    # Set the font size of the tick labels.
    for tick in ax[0].yaxis.get_major_ticks():
        tick.label.set_fontsize(tick_fontsize)
    for tick in ax[1].xaxis.get_major_ticks():
        tick.label.set_fontsize(tick_fontsize)
    for tick in ax[1].yaxis.get_major_ticks():
        tick.label.set_fontsize(tick_fontsize)

    plt.tight_layout()

    if rethist:
        return (fig, hist, xedges, yedges)

    return fig



# -*- coding: utf-8 -*-

"""The llhratio module provides classes implementing the log-likelihood ratio
functions. In general these should be detector independent, because they
implement the pure math of the log-likelihood ratio function.
"""

import abc
import numpy as np

from skyllh.core.config import (
    HasConfig,
)
from skyllh.core.debugging import (
    get_logger,
)
from skyllh.core.minimizer import (
    Minimizer,
    NR1dNsMinimizerImpl,
    NRNsScan2dMinimizerImpl,
)
from skyllh.core.parameters import (
    ParameterModelMapper,
)
from skyllh.core.pdfratio import (
    PDFRatio,
)
from skyllh.core.py import (
    classname,
    issequenceof,
    float_cast,
)
from skyllh.core.services import (
    DatasetSignalWeightFactorsService,
    SrcDetSigYieldWeightsService,
)
from skyllh.core.source_hypo_grouping import (
    SourceHypoGroupManager,
)
from skyllh.core.timing import (
    TaskTimer,
)
from skyllh.core.trialdata import (
    TrialDataManager,
)


logger = get_logger(__name__)


class LLHRatio(
        HasConfig,
        metaclass=abc.ABCMeta,
):
    """Abstract base class for a log-likelihood (LLH) ratio function.
    """

    def __init__(
            self,
            pmm,
            minimizer,
            **kwargs):
        """Creates a new LLH ratio function instance.

        Parameters
        ----------
        pmm : instance of ParameterModelMapper
            The instance of ParameterModelMapper providing the mapping of
            global parameters to local parameters of individual models.
        minimizer : instance of Minimizer
            The Minimizer instance that should be used to minimize the negative
            of this log-likelihood ratio function.
        """
        super().__init__(**kwargs)

        self.pmm = pmm
        self.minimizer = minimizer

    @property
    def pmm(self):
        """The ParameterModelMapper instance providing the mapping of
        global floating parameters to individual models.
        """
        return self._pmm

    @pmm.setter
    def pmm(self, mapper):
        if not isinstance(mapper, ParameterModelMapper):
            raise TypeError(
                'The pmm property must be an instance of '
                'ParameterModelMapper! '
                f'Its current type is {classname(mapper)}.')
        self._pmm = mapper

    @property
    def minimizer(self):
        """The Minimizer instance used to minimize the negative of the
        log-likelihood ratio function.
        """
        return self._minimizer

    @minimizer.setter
    def minimizer(self, minimizer):
        if not isinstance(minimizer, Minimizer):
            raise TypeError(
                'The minimizer property must be an instance of Minimizer! '
                f'Its current type is {classname(minimizer)}.')
        self._minimizer = minimizer

    @abc.abstractmethod
    def initialize_for_new_trial(
            self,
            tl=None,
            **kwargs):
        """This method will be called by the Analysis class after new trial data
        has been initialized to the trial data manager. Derived classes can make
        use of this call hook to perform LLHRatio specific trial initialization.

        Parameters
        ----------
        tl : instance of TimeLord | None
            The optional instance of TimeLord to use for timing measurements.
        """
        pass

    @abc.abstractmethod
    def evaluate(
            self,
            fitparam_values,
            src_params_recarray=None,
            tl=None):
        """This method evaluates the LLH ratio function for the given set of
        fit parameter values.

        Parameters
        ----------
        fitparam_values : instance of numpy ndarray
            The (N_fitparams,)-shaped numpy 1D ndarray holding the current
            values of the global fit parameters.
        src_params_recarray : instance of numpy record ndarray | None
            The numpy record ndarray of length N_sources holding the parameter
            names and values of all sources. If set to ``None`` it will be
            created automatically from the ``fitparam_values`` array.
            See the documentation of the
            :meth:`skyllh.core.parameters.ParameterModelMapper.create_src_params_recarray`
            method for more information about this array.
        tl : instance of TimeLord | None
            The optional instance of TimeLord to use for measuring timing.

        Returns
        -------
        log_lambda : float
            The calculated log-lambda value.
        grads : instance of numpy ndarray
            The (N_fitparams,)-shaped 1D numpy ndarray holding the gradient
            value for each global fit parameter.
        """
        pass

    def maximize(
            self,
            rss,
            tl=None):
        """Maximize the log-likelihood ratio function, by using the ``evaluate``
        method.

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService to draw random numbers from.
            This is needed to generate random parameter initial values.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to time the
            maximization process.

        Returns
        -------
        log_lambda_max : float
            The (maximum) value of the log-likelihood ratio (log_lambda)
            function for the best fit parameter values.
        fitparam_values : instance of numpy ndarray
            The (N_fitparam,)-shaped 1D numpy ndarray holding the values of the
            global fit parameters.
        status : dict
            The dictionary with status information about the maximization
            process, i.e. from the minimizer.
        """
        tracing = self._cfg['debugging']['enable_tracing']

        # Define the negative llhratio function, that will get minimized.
        self_evaluate = self.evaluate

        def negative_llhratio_func(fitparam_values, func_stats, tl=None):
            src_params_recarray = self._pmm.create_src_params_recarray(
                fitparam_values)

            func_stats['n_calls'] += 1
            with TaskTimer(tl, 'Evaluate llh-ratio function.'):
                (f, grads) = self_evaluate(
                    fitparam_values=fitparam_values,
                    src_params_recarray=src_params_recarray,
                    tl=tl)
                if tracing:
                    logger.debug(
                        f'LLH-ratio func value f={f:g}, grads={str(grads)}')
            return (-f, -grads)

        minimize_kwargs = {
            'func_provides_grads': True
        }

        func_stats = {
            'n_calls': 0
        }
        with TaskTimer(tl, 'Minimize -llhratio function.'):
            (fitparam_values, fmin, status) = self._minimizer.minimize(
                rss=rss,
                paramset=self._pmm.global_paramset,
                func=negative_llhratio_func,
                args=(func_stats, tl),
                kwargs=minimize_kwargs)
        log_lambda_max = -fmin
        status['n_llhratio_func_calls'] = func_stats['n_calls']

        logger.debug(
            f'Maximized LLH ratio function with '
            f'{status["n_llhratio_func_calls"]:d} calls')

        return (log_lambda_max, fitparam_values, status)


class TCLLHRatio(
        LLHRatio,
        metaclass=abc.ABCMeta):
    """Abstract base class for a log-likelihood (LLH) ratio function with two
    components, i.e. signal and background.
    """

    def __init__(
            self,
            pmm,
            minimizer,
            mean_n_sig_0,
            **kwargs):
        """Creates a new two-component LLH ratio function instance.

        Parameters
        ----------
        pmm : instance of ParameterModelMapper
            The instance of ParameterModelMapper providing the mapping of
            global floating parameters to individual models.
        minimizer : instance of Minimizer
            The Minimizer instance that should be used to minimize the negative
            of this log-likelihood ratio function.
        mean_n_sig_0 : float
            The fixed mean number of signal events for the null-hypothesis.
        """
        super().__init__(
            pmm=pmm,
            minimizer=minimizer,
            **kwargs)

        self.mean_n_sig_0 = mean_n_sig_0

    @property
    def mean_n_sig_0(self):
        """The parameter value for the mean number of signal events of the
        null-hypothesis.
        """
        return self._mean_n_sig_0

    @mean_n_sig_0.setter
    def mean_n_sig_0(self, v):
        v = float_cast(
            v,
            'The mean_n_sig_0 property must be cast-able to a float value!')
        self._mean_n_sig_0 = v

    @abc.abstractmethod
    def calculate_ns_grad2(
            self,
            ns,
            ns_pidx,
            src_params_recarray,
            tl=None,
            **kwargs):
        """This method is supposed to calculate the second derivative of the
        log-likelihood ratio function w.r.t. the fit parameter ns, the number
        of signal events in the data set.

        Parameters
        ----------
        fitparam_values : instance of numpy ndarray
            The (N_fitparams,)-shaped 1D numpy ndarray holding the current
            values of the global fit parameters.
        ns_pidx : int
            The index of the global ns fit parameter.
        src_params_recarray : instance of numpy record ndarray
            The numpy record ndarray of length N_sources holding the parameter
            names and values of all sources.
            See the documentation of the
            :meth:`skyllh.core.parameters.ParameterModelMapper.create_src_params_recarray`
            method for more information about this array.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used for timing
            measurements.

        Returns
        -------
        nsgrad2 : float
            The second derivative w.r.t. ns of the log-likelihood ratio function
            for the given fit parameter values.
        """
        pass

    def maximize_with_1d_newton_rapson_minimizer(
            self,
            rss,
            tl=None):
        """Maximizes this log-likelihood ratio function, by minimizing its
        negative using a 1D Newton-Rapson minimizer.

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService that should be used to draw
            random numbers from. It is used by the minimizer to generate random
            fit parameter initial values.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to time the
            maximization of the LLH-ratio function.

        Returns
        -------
        log_lambda_max : float
            The (maximum) value of the log-likelihood ratio (log_lambda)
            function for the best fit parameter values.
        fitparam_values : instance of numpy ndarray
            The (N_fitparam,)-shaped 1D numpy ndarray holding the global
            fit parameter values.
        status : dict
            The dictionary with status information about the maximization
            process, i.e. from the minimizer.
        """
        # Define the negative llhratio function, that will get minimized
        # when using the Newton-Rapson 1D minimizer for llhratio functions
        # depending solely on ns.
        self__evaluate = self.evaluate
        self__calculate_ns_grad2 = self.calculate_ns_grad2

        ns_pidx = self._pmm.get_gflp_idx(name='ns')

        def negative_llhratio_func_nr1d_ns(fitparam_values, tl):
            ns = fitparam_values[ns_pidx]
            src_params_recarray = self._pmm.create_src_params_recarray(
                fitparam_values)
            with TaskTimer(
                    tl,
                    'Evaluate llh-ratio function.'):
                (f, grads) = self__evaluate(
                    fitparam_values=fitparam_values,
                    src_params_recarray=src_params_recarray,
                    tl=tl)
            with TaskTimer(
                    tl,
                    'Calculate 2nd derivative of llh-ratio function w.r.t. ns'):
                grad2_ns = self__calculate_ns_grad2(
                    ns=ns,
                    ns_pidx=ns_pidx,
                    src_params_recarray=src_params_recarray,
                    tl=tl)

            return (-f, -grads[ns_pidx], -grad2_ns)

        (fitparam_values, fmin, status) = self._minimizer.minimize(
            rss=rss,
            paramset=self._pmm.global_paramset,
            func=negative_llhratio_func_nr1d_ns,
            args=(tl,))
        log_lambda_max = -fmin

        return (log_lambda_max, fitparam_values, status)

    def maximize(
            self,
            rss,
            tl=None):
        """Maximizes this log-likelihood ratio function, by minimizing its
        negative.
        This method has a special implementation when a 1D Newton-Rapson
        minimizer is used. In that case only the first and second derivative
        of the log-likelihood ratio function is calculated.

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService that should be used to draw
            random numbers from. It is used by the minimizer to generate random
            fit parameter initial values.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to time the
            maximization of the LLH-ratio function.

        Returns
        -------
        log_lambda_max : float
            The (maximum) value of the log-likelihood ratio (log_lambda)
            function for the best fit parameter values.
        fitparam_values : instance of numpy ndarray
            The (N_fitparam,)-shaped 1D numpy ndarray holding the global
            fit parameter values.
        status : dict
            The dictionary with status information about the maximization
            process, i.e. from the minimizer.
        """
        if isinstance(self._minimizer.minimizer_impl, NR1dNsMinimizerImpl) or\
           isinstance(self._minimizer.minimizer_impl, NRNsScan2dMinimizerImpl):
            return self.maximize_with_1d_newton_rapson_minimizer(
                rss=rss,
                tl=tl)

        return super().maximize(
            rss=rss,
            tl=tl)


class SingleDatasetTCLLHRatio(
        TCLLHRatio,
        metaclass=abc.ABCMeta):
    """Abstract base class for a log-likelihood (LLH) ratio function with two
    components, i.e. signal and background, for a single data set.
    """

    def __init__(
            self,
            pmm,
            minimizer,
            shg_mgr,
            tdm,
            mean_n_sig_0,
            **kwargs):
        """Creates a new two-component LLH ratio function instance for a single
        data set.

        Parameters
        ----------
        pmm : instance of ParameterModelMapper
            The instance of ParameterModelMapper providing the mapping of
            global floating parameters to individual models.
        minimizer : instance of Minimizer
            The Minimizer instance that should be used to minimize the negative
            of this log-likelihood ratio function.
        shg_mgr : SourceHypoGroupManager instance
            The SourceHypoGroupManager instance that defines the source
            hypothesis groups.
        tdm : instance of TrialDataManager
            The instance of TrialDataManager that holds the trial event data and
            additional data fields for this LLH ratio function.
        mean_n_sig_0 : float
            The fixed mean number of signal events for the null-hypothesis.
        """
        super().__init__(
            pmm=pmm,
            minimizer=minimizer,
            mean_n_sig_0=mean_n_sig_0,
            **kwargs)

        self.shg_mgr = shg_mgr
        self.tdm = tdm

        # Calculate the data fields that solely depend on source parameters.
        self._tdm.calculate_source_data_fields(
            shg_mgr=self._shg_mgr,
            pmm=self._pmm)

    @property
    def shg_mgr(self):
        """The SourceHypoGroupManager instance that defines the source
        hypothesis groups.
        """
        return self._shg_mgr

    @shg_mgr.setter
    def shg_mgr(self, mgr):
        if not isinstance(mgr, SourceHypoGroupManager):
            raise TypeError(
                'The shg_mgr property must be an instance of '
                'SourceHypoGroupManager! '
                f'Its current type is {classname(mgr)}.')
        self._shg_mgr = mgr

    @property
    def tdm(self):
        """The instance of TrialDataManager that holds the trial event data and
        additional data fields for this LLH ratio function.
        """
        return self._tdm

    @tdm.setter
    def tdm(self, mgr):
        if not isinstance(mgr, TrialDataManager):
            raise TypeError(
                'The tdm property must be an instance of TrialDataManager! '
                f'Its current type is {classname(mgr)}.')
        self._tdm = mgr

    def change_shg_mgr(self, shg_mgr):
        """Changes the source hypothesis group manager of this two-component LLH
        ratio function.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The new instance of SourceHypoGroupManager.
        """
        self.shg_mgr = shg_mgr

        self._tdm.change_shg_mgr(
            shg_mgr=shg_mgr,
            pmm=self._pmm)


class ZeroSigH0SingleDatasetTCLLHRatio(
        SingleDatasetTCLLHRatio):
    """This class implements a two-component (TC) log-likelihood (LLH) ratio
    function for a single dataset assuming zero signal for the null-hypothesis.
    """
    # The (1 + alpha)-threshold float value for which the log-likelihood ratio
    # function of a single event should get approximated by a Taylor expansion.
    # This is to prevent a divergence of the log-function for each event, where
    # (1 + alpha_i) < (1 + alpha).
    # This setting is implemented as a class type member instead of a class
    # instance member, because it is supposed to be the same for all instances.
    _one_plus_alpha = 1e-3

    def __init__(
            self,
            pmm,
            minimizer,
            shg_mgr,
            tdm,
            pdfratio,
            **kwargs):
        """Constructor of the two-component log-likelihood ratio function.

        Parameters
        ----------
        pmm : instance of ParameterModelMapper
            The instance of ParameterModelMapper providing the mapping of
            global floating parameters to individual models.
        minimizer : instance of Minimizer
            The Minimizer instance that should be used to minimize the negative
            of this log-likelihood ratio function.
        shg_mgr : instance of SourceHypoGroupManager
            The SourceHypoGroupManager instance that defines the source
            hypothesis groups.
        tdm : instance of TrialDataManager
            The instance of TrialDataManager that holds the trial event data and
            additional data fields for this LLH ratio function.
        pdfratio : instance of PDFRatio
            The instance of PDFRatio. A PDFRatio instance might depend
            on none, one, or several fit parameters.
        """
        super().__init__(
            pmm=pmm,
            minimizer=minimizer,
            shg_mgr=shg_mgr,
            tdm=tdm,
            mean_n_sig_0=0,
            **kwargs)

        self.pdfratio = pdfratio

        # Define cache variable for the evaluate method to store values needed
        # for a possible calculation of the second derivative w.r.t. ns of the
        # log-likelihood ratio function.
        self._cache_nsgrad_i = None

    @SingleDatasetTCLLHRatio.mean_n_sig_0.setter
    def mean_n_sig_0(self, v):
        SingleDatasetTCLLHRatio.mean_n_sig_0.fset(self, v)
        if self._mean_n_sig_0 != 0:
            raise ValueError(
                f'The {classname(self)} class is only valid for '
                f'mean_n_sig_0 = 0!')

    @property
    def pdfratio(self):
        """The instance of PDFRatio.
        """
        return self._pdfratio

    @pdfratio.setter
    def pdfratio(self, r):
        if not isinstance(r, PDFRatio):
            raise TypeError(
                'The pdfratio property must be an instance of PDFRatio! '
                f'Its current type is {classname(r)}.')
        self._pdfratio = r

    def initialize_for_new_trial(
            self,
            tl=None,
            **kwargs):
        """Initializes the log-likelihood ratio function for a new trial.
        It calls the
        :meth:`~skyllh.core.pdfratio.PDFRatio.initialize_for_new_trial` method
        of the :class:`~skyllh.core.pdfratio.PDFRatio` class.

        Parameters
        ----------
        tl : instance of TimeLord
            The optional instance of TimeLord to measure timing information.
        """
        self._pdfratio.initialize_for_new_trial(
            tdm=self._tdm,
            tl=tl,
            **kwargs)

    def calculate_log_lambda_and_grads(
            self,
            N,
            ns,
            ns_pidx,
            p_mask,
            Xi,
            dXi_dp):
        """Calculates the log(Lambda) value and its gradient for each global fit
        parameter. This calculation is source and detector independent.

        Parameters
        ----------
        fitparam_values : instance of numpy ndarray
            The (N_fitparams,)-shaped ndarray holding the current values of the
            global fit parameters.
            These numbers are used as cache key to validate the ``nsgrad_i``
            values for the given fit parameter values for a possible later
            calculation of the second derivative w.r.t. ns of the log-likelihood
            ratio function.
        N : int
            The total number of events.
        ns : float
            The value of the global fit parameter ns.
        ns_pidx : int
            The index of the global fit parameter ns.
        p_mask : instance of numpy ndarray
            The (N_fitparam,)-shaped numpy ndarray of bool selecting all global
            fit parameters, except ns.
        Xi : instance of numpy ndarray
            The (n_selected_events,)-shaped 1D numpy ndarray holding the X value
            of each selected event.
        dXi_dp : instance of numpy ndarray
            The (n_selected_events, N_fitparams-1,)-shaped 2D ndarray holding
            the derivative value for each fit parameter p (i.e. except ns) of
            each event's X value.

        Returns
        -------
        log_lambda : float
            The value of the log-likelihood ratio function.
        grads : instance of numpy ndarray
            The (N_fitparams,)-shaped numpy ndarray holding the gradient value
            of log_lambda for each fit parameter.
        """
        tracing = self._cfg['debugging']['enable_tracing']

        # Get the number of selected events.
        Nprime = len(Xi)

        if tracing:
            logger.debug(
                f'N={N:d}, Nprime={Nprime:d}')

        one_plus_alpha = ZeroSigH0SingleDatasetTCLLHRatio._one_plus_alpha

        alpha = one_plus_alpha - 1
        alpha_i = ns*Xi

        # Create a mask for events which have a stable non-diverging
        # log-function argument, and an inverted mask thereof.
        m_stable = alpha_i > alpha
        m_unstable = ~m_stable
        any_unstable_events = np.any(m_unstable)

        if tracing:
            logger.debug(
                '# of events doing Taylor expansion for (unstable events): '
                f'{np.count_nonzero(m_unstable):d}')

        # Allocate memory for the log_lambda_i values.
        log_lambda_i = np.empty_like(alpha_i, dtype=np.float64)

        # Calculate the log_lambda_i value for the numerical stable events.
        np.log1p(alpha_i, where=m_stable, out=log_lambda_i)

        # Calculate the log_lambda_i value for the numerical unstable events.
        if any_unstable_events:
            tildealpha_i = (alpha_i[m_unstable] - alpha) / one_plus_alpha
            log_lambda_i[m_unstable] =\
                np.log1p(alpha) + tildealpha_i - 0.5 * tildealpha_i**2

        # Calculate the log_lambda value and account for pure background events.
        log_lambda = np.sum(log_lambda_i) + (N - Nprime)*np.log1p(-ns/N)

        # Calculate the gradient for each fit parameter.
        grads = np.empty((dXi_dp.shape[1]+1,), dtype=np.float64)

        # Pre-calculate value that is used twice for the gradients of the
        # numerical stable events.
        one_over_one_plus_alpha_i_stablemask = 1 / (1 + alpha_i[m_stable])

        # For ns.
        nsgrad_i = np.empty_like(alpha_i, dtype=np.float64)
        nsgrad_i[m_stable] =\
            Xi[m_stable] * one_over_one_plus_alpha_i_stablemask
        if any_unstable_events:
            nsgrad_i[m_unstable] =\
                (1 - tildealpha_i) * Xi[m_unstable] / one_plus_alpha

        # Cache the nsgrad_i values for a possible later calculation of the
        # second derivative w.r.t. ns of the log-likelihood ratio function.
        self._cache_nsgrad_i = nsgrad_i

        # Calculate the first derivative w.r.t. ns.
        grads[ns_pidx] = np.sum(nsgrad_i) - (N - Nprime) / (N - ns)

        # Now for each other fit parameter.

        # For all numerical stable events.
        grads[p_mask] = np.sum(
            ns * one_over_one_plus_alpha_i_stablemask[:, np.newaxis] *
            dXi_dp[m_stable],
            axis=0)

        # For all numerical unstable events.
        if any_unstable_events:
            grads[p_mask] += np.sum(
                ns * (1 - tildealpha_i[:, np.newaxis]) * dXi_dp[m_unstable] /
                one_plus_alpha,
                axis=0)

        return (log_lambda, grads)

    def calculate_ns_grad2(
            self,
            ns,
            ns_pidx=None,
            src_params_recarray=None,
            tl=None):
        """Calculates the second derivative w.r.t. ns of the log-likelihood
        ratio function.
        This method tries to use cached values for the first derivative
        w.r.t. ns of the log-likelihood ratio function for the individual
        events. If cached values don't exist or do not match the given fit
        parameter values, they will get calculated automatically by calling the
        evaluate method with the given fit parameter values.

        Parameters
        ----------
        fitparam_values : numpy (N_fitparams+1)-shaped 1D ndarray
            The ndarray holding the current values of the global fit
            parameters.
        ns : float
            The value of the global fit parameter ns.
        ns_pidx : int
            The parameter index of the global fit parameter ns.
            For this particular class this is an ignored interface parameter.
        src_params_recarray : instance of numpy record ndarray
            The numpy record ndarray of length N_sources holding the parameter
            names and values of all sources.
            See the documentation of the
            :meth:`skyllh.core.parameters.ParameterModelMapper.create_src_params_recarray`
            method for more information about this array.
            For this particular class this is an ignored interface parameter.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used for timing
            measurements.

        Returns
        -------
        nsgrad2 : float
            The second derivative w.r.t. ns of the log-likelihood ratio function
            for the given fit parameter values.
        """
        if self._cache_nsgrad_i is None:
            raise RuntimeError(
                'The evaluate method needs to be called before the '
                'calculate_ns_grad2 method can be called!')

        Nprime = self._tdm.n_selected_events
        N = Nprime + self._tdm.n_pure_bkg_events

        nsgrad2 = -np.sum(self._cache_nsgrad_i**2) - (N - Nprime)/(N - ns)**2

        return nsgrad2

    def evaluate(
            self,
            fitparam_values,
            src_params_recarray=None,
            tl=None):
        """Evaluates the log-likelihood ratio function for the given set of
        data events.

        Parameters
        ----------
        fitparam_values : instance of numpy ndarray
            The (N_fitparams,)-shaped 1D ndarray holding the current values of
            the global fit parameters.
        src_params_recarray : instance of numpy structured ndarray | None
            The numpy record ndarray of length N_sources holding the local
            parameter names and values of all sources.
            If it is ``None``, it will be generated automatically from the
            ``fitparam_values`` argument using the
            :class:`~skyllh.core.parameters.ParameterModelMapper` instance.
        tl : instance of TimeLord | None
            The optional instance of TimeLord to measure the timing of
            evaluating the LLH ratio function.

        Returns
        -------
        log_lambda : float
            The calculated log-lambda value.
        grads : instance of numpy ndarray
            The (N_fitparams,)-shaped 1D numpy ndarray holding the gradient
            value for each global fit parameter.
        """
        tracing = self._cfg['debugging']['enable_tracing']

        if src_params_recarray is None:
            src_params_recarray = self._pmm.create_src_params_recarray(
                gflp_values=fitparam_values
            )

        tdm = self._tdm

        ns_pidx = self._pmm.get_gflp_idx('ns')

        ns = fitparam_values[ns_pidx]

        N = tdm.n_events

        # Calculate the data fields that depend on global fit parameters.
        if tdm.has_global_fitparam_data_fields:
            with TaskTimer(
                    tl,
                    'Calculate global fit parameter dependent data fields.'):
                # Create the global_fitparams dictionary with the global fit
                # parameter names and values.
                global_fitparams = self._pmm.get_global_floating_params_dict(
                    gflp_values=fitparam_values)
                tdm.calculate_global_fitparam_data_fields(
                    shg_mgr=self._shg_mgr,
                    pmm=self._pmm,
                    global_fitparams=global_fitparams)

        # Calculate the PDF ratio values for each selected event.
        with TaskTimer(tl, 'Calc pdfratio value Ri'):
            Ri = self._pdfratio.get_ratio(
                tdm=tdm,
                src_params_recarray=src_params_recarray,
                tl=tl)

        # Calculate Xi for each selected event.
        Xi = (Ri - 1.) / N

        n_fitparams = len(fitparam_values)

        # Calculate the gradients of Xi for each fit parameter (without ns).
        dXi_dp = np.empty(
            (Xi.shape[0], n_fitparams-1),
            dtype=np.float64)

        # Create a mask that selects all fit parameters except ns.
        p_mask = np.ones((n_fitparams,), dtype=np.bool_)
        p_mask[ns_pidx] = False

        # Loop over the global fit parameters and calculate the derivative of
        # Xi w.r.t. each fit parameter.
        fitparam_ids = np.arange(n_fitparams)
        for (idx, fitparam_id) in enumerate(fitparam_ids[p_mask]):
            dRi = self._pdfratio.get_gradient(
                tdm=tdm,
                src_params_recarray=src_params_recarray,
                fitparam_id=fitparam_id,
                tl=tl)

            # Calculate the derivative of Xi w.r.t. the global fit parameter
            # with ID fitparam_id.
            dXi_dp[:, idx] = dRi / N

        if tracing:
            logger.debug(
                f'{classname(self)}.evaluate: N={N}, Nprime={len(Xi)}, '
                f'ns={ns:.3f}')

        with TaskTimer(tl, 'Calc logLambda and grads'):
            (log_lambda, grads) = self.calculate_log_lambda_and_grads(
                N=N,
                ns=ns,
                ns_pidx=ns_pidx,
                p_mask=p_mask,
                Xi=Xi,
                dXi_dp=dXi_dp)

        return (log_lambda, grads)


class MultiDatasetTCLLHRatio(
        TCLLHRatio):
    """This class describes a two-component log-likelihood ratio function for
    multiple datasets. The final log-likelihood ratio value is the sum of the
    individual log-likelihood ratio values.

    The different datasets contribute according to their dataset signal weight
    factor, f_j(p_s), which depends on possible signal fit parameters. By
    definition the signal fit parameters are assumed to be the same for each
    dataset.

    By mathematical definition this class is suitable for single and multi
    source hypotheses.
    """
    def __init__(
            self,
            pmm,
            minimizer,
            src_detsigyield_weights_service,
            ds_sig_weight_factors_service,
            llhratio_list,
            **kwargs):
        """Creates a new composite two-component log-likelihood ratio function.

        Parameters
        ----------
        pmm : instance of ParameterModelMapper
            The instance of ParameterModelMapper providing the mapping of
            global floating parameters to individual models.
        minimizer : instance of Minimizer
            The Minimizer instance that should be used to minimize the negative
            of this log-likelihood ratio function.
        src_detsigyield_weights_service : instance of SrcDetSigYieldWeightsService
            An instance of SrcDetSigYieldWeightsService, which provides the
            product of the source weights with the detector signal yield.
        ds_sig_weight_factors_service : instance of DatasetSignalWeightFactorsService
            An instance of DatasetSignalWeightFactorsService, which provides
            the relative dataset signal weight factors.
        llhratio_list : list of instance of SingleDatasetTCLLHRatio
            The list of the two-component log-likelihood ratio functions,
            one for each dataset.
        """
        if not issequenceof(llhratio_list, SingleDatasetTCLLHRatio):
            raise TypeError(
                'The llhratio_list argument must be a sequence of '
                'SingleDatasetTCLLHRatio instances! '
                f'Its current type is {classname(llhratio_list)}.')
        self._llhratio_list = list(llhratio_list)

        super().__init__(
            pmm=pmm,
            minimizer=minimizer,
            mean_n_sig_0=self._llhratio_list[0].mean_n_sig_0,
            **kwargs)

        self.src_detsigyield_weights_service = src_detsigyield_weights_service
        self.ds_sig_weight_factors_service = ds_sig_weight_factors_service

        if (
            self.ds_sig_weight_factors_service.n_datasets
            != len(self._llhratio_list)
           ):
            raise ValueError(
                'The number of datasets the DatasetSignalWeightFactorsService '
                'instance is made for must be equal to the number of '
                'log-likelihood ratio functions!')

    @property
    def src_detsigyield_weights_service(self):
        """The instance of SrcDetSigYieldWeightsService, which provides the
        product of the source weights with the detector signal yield.
        """
        return self._src_detsigyield_weights_service

    @src_detsigyield_weights_service.setter
    def src_detsigyield_weights_service(self, service):
        if not isinstance(service, SrcDetSigYieldWeightsService):
            raise TypeError(
                'The src_detsigyield_weights_service property must be an '
                'instance of SrcDetSigYieldWeightsService! '
                f'Its current type is {classname(service)}.')
        self._src_detsigyield_weights_service = service

    @property
    def ds_sig_weight_factors_service(self):
        """The instance of DatasetSignalWeightFactorsService that provides the
        relative dataset signal weight factors.
        """
        return self._ds_sig_weight_factors_service

    @ds_sig_weight_factors_service.setter
    def ds_sig_weight_factors_service(self, service):
        if not isinstance(service, DatasetSignalWeightFactorsService):
            raise TypeError(
                'The ds_sig_weight_factors_service property must be an '
                'instance of DatasetSignalWeightFactorsService! '
                f'Its current type is {classname(service)}.')
        self._ds_sig_weight_factors_service = service

    @property
    def llhratio_list(self):
        """(read-only) The list of TCLLHRatio instances, which are part of this
        composite log-likelihood-ratio function.
        """
        return self._llhratio_list

    @llhratio_list.setter
    def llhratio_list(self, llhratios):
        if not issequenceof(llhratios, SingleDatasetTCLLHRatio):
            raise TypeError(
                'The llhratio_list property must be a sequence of '
                'SingleDatasetTCLLHRatio instances! '
                f'Its current type is {classname(llhratios)}.')
        self._llhratio_list = list(llhratios)

    @property
    def n_selected_events(self):
        """(read-only) The sum of selected events of each individual
        log-likelihood ratio function.
        """
        n_selected_events = 0
        for llhratio in self._llhratio_list:
            n_selected_events += llhratio.tdm.n_selected_events
        return n_selected_events

    @TCLLHRatio.mean_n_sig_0.setter
    def mean_n_sig_0(self, v):
        TCLLHRatio.mean_n_sig_0.fset(self, v)
        for llhratio in self._llhratio_list:
            llhratio.mean_n_sig_0 = self._mean_n_sig_0

    def change_shg_mgr(self, shg_mgr):
        """Changes the source hypo group manager of all objects of this LLH
        ratio function, hence, calling the ``change_shg_mgr``
        method of all TCLLHRatio instances of this LLHRatio instance.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager that defines the groups of
            source hypotheses.
        """
        self._src_detsigyield_weights_service.change_shg_mgr(
            shg_mgr=shg_mgr)

        for llhratio in self._llhratio_list:
            llhratio.change_shg_mgr(
                shg_mgr=shg_mgr)

    def initialize_for_new_trial(
            self,
            tl=None,
            **kwargs):
        """Initializes the log-likelihood-ratio function for a new trial.
        It calls the
        :meth:`~skyllh.core.llhratio.LLHRatio.initialize_for_new_trial` method
        of the :class:`~skyllh.core.llhratio.LLHRatio` class of each individual
        log-likelihood ratio function.

        Parameters
        ----------
        tl : instance of TimeLord
            The optional instance of TimeLord to measure timing information.
        """
        for llhratio in self._llhratio_list:
            llhratio.initialize_for_new_trial(
                tl=tl,
                **kwargs)

    def evaluate(
            self,
            fitparam_values,
            src_params_recarray=None,
            tl=None):
        """Evaluates the composite log-likelihood-ratio function and returns its
        value and global fit parameter gradients.

        Parameters
        ----------
        fitparam_values : instance of numpy ndarray
            The (N_fitparams,)-shaped numpy 1D ndarray holding the current
            values of the global fit parameters.
        src_params_recarray : instance of numpy record ndarray | None
            The numpy record ndarray of length N_sources holding the parameter
            names and values of all sources.
            See the documentation of the
            :meth:`skyllh.core.parameters.ParameterModelMapper.create_src_params_recarray`
            method for more information about this array.
            It case it is ``None``, it will be created automatically from the
            ``fitparam_values`` argument using the
            :class:`~skyllh.core.parameters.ParameterModelMapper` instance.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used for timing
            measurements.

        Returns
        -------
        log_lambda : float
            The calculated log-lambda value of the composite
            log-likelihood-ratio function.
        grads : instance of numpy ndarray
            The (N_fitparams,)-shaped 1D ndarray holding the gradient value of
            the composite log-likelihood-ratio function for each global fit
            parameter.
        """
        tracing = self._cfg['debugging']['enable_tracing']

        if src_params_recarray is None:
            src_params_recarray = self._pmm.create_src_params_recarray(
                gflp_values=fitparam_values
            )

        n_fitparams = len(fitparam_values)

        ns_pidx = self._pmm.get_gflp_idx('ns')

        ns = fitparam_values[ns_pidx]
        if tracing:
            logger.debug(
                f'{classname(self)}.evaluate: ns={ns:.3f}')

        # We need to calculate the source detsigyield weights and the dataset
        # signal weight factors.
        self._src_detsigyield_weights_service.calculate(
            src_params_recarray=src_params_recarray)
        self._ds_sig_weight_factors_service.calculate()

        # Get the dataset signal weights and their gradients.
        # f is a (N_datasets,)-shaped 1D ndarray.
        # f_grads is a dictionary holding (N_datasets,)-shaped 1D ndarrays for
        # each global fit parameter.
        (f, f_grads_dict) = self._ds_sig_weight_factors_service.get_weights()

        # Convert the f_grads dictionary into a (N_datasets,N_fitparams)
        f_grads = np.zeros((len(f), n_fitparams), dtype=np.float64)
        for pidx in f_grads_dict.keys():
            f_grads[:, pidx] = f_grads_dict[pidx]

        nsf = ns * f

        # Calculate the composite log-likelihood-ratio function value and the
        # gradient of the composite log-likelihood ratio function for each
        # global fit parameter.
        log_lambda = 0

        # Allocate an array for the gradients of the composite log-likelihood
        # function. It is always at least one element long, i.e. the gradient
        # for ns.
        grads = np.zeros((n_fitparams,), dtype=np.float64)

        # Create an array holding the fit parameter values for a particular
        # llh ratio function. Since we need to adjust ns with nsj it's more
        # efficient to create this array once and use it within the for loop
        # over the llh ratio functions.
        llhratio_fitparam_values = fitparam_values.copy()

        pmask = np.ones((n_fitparams,), dtype=np.bool_)
        pmask[ns_pidx] = False

        # Loop over the llh ratio functions.
        for (j, llhratio) in enumerate(self._llhratio_list):
            if tracing:
                logger.debug(
                    f'nsf[j={j}] = {nsf[j]:.3f}')

            llhratio_fitparam_values[ns_pidx] = nsf[j]

            (log_lambda_j, grads_j) = llhratio.evaluate(
                fitparam_values=llhratio_fitparam_values,
                src_params_recarray=src_params_recarray,
                tl=tl)
            log_lambda += log_lambda_j

            # Gradient for ns.
            grads[ns_pidx] += grads_j[ns_pidx] * f[j]

            # Gradient for each global fit parameter, if there are any.
            if len(grads) > 1:
                ns_summand = grads_j[ns_pidx] * ns * f_grads[j][pmask]
                grads[pmask] += ns_summand + grads_j[pmask]

        return (log_lambda, grads)

    def calculate_ns_grad2(
            self,
            ns,
            ns_pidx,
            src_params_recarray,
            tl=None):
        """Calculates the second derivative w.r.t. ns of the log-likelihood
        ratio function.

        Note::

            This method takes the dataset signal weight factors from the dataset
            signal weight factors service. Hence, the service needs to be
            updated before calling this method.

        Parameters
        ----------
        fitparam_values : instance of numpy ndarray
            The (N_fitparams,)-shaped 1D ndarray holding the current values of
            the global fit parameters.
        ns : float
            The value of the global fit parameter ns.
        ns_pidx : int
            The index of the global parameter ns.
        src_params_recarray : instance of numpy record ndarray
            The numpy record ndarray of length N_sources holding the parameter
            names and values of all sources.
            See the documentation of the
            :meth:`skyllh.core.parameters.ParameterModelMapper.create_src_params_recarray`
            method for more information about this array.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used for timing
            measurements.

        Returns
        -------
        nsgrad2 : float
            The second derivative w.r.t. ns of the log-likelihood ratio function
            for the given fit parameter values.
        """
        (f, f_grads_dict) = self._ds_sig_weight_factors_service.get_weights()

        nsf = ns * f

        nsgrad2j = np.empty((len(self._llhratio_list),), dtype=np.float64)

        # Loop over the llh ratio functions and calculate their second
        # derivative.
        for (j, llhratio) in enumerate(self._llhratio_list):
            nsgrad2j[j] = llhratio.calculate_ns_grad2(
                ns=nsf[j],
                ns_pidx=ns_pidx,
                src_params_recarray=src_params_recarray,
                tl=tl)

        nsgrad2 = np.sum(nsgrad2j * f**2)

        return nsgrad2


class NsProfileMultiDatasetTCLLHRatio(
        TCLLHRatio):
    r"""This class implements a profile log-likelihood ratio function that has
    only ns as fit parameter. It uses a MultiDatasetTCLLHRatio instance as
    log-likelihood function. Hence, mathematically it is

    .. math::

        \Lambda(n_{\mathrm{s}}) = \frac{L(n_{\mathrm{s}})}{L(n_{\mathrm{s}}=n_{\mathrm{s},0})},

    where :math:`n_{\mathrm{s},0}` is the fixed mean number of signal events for
    the null-hypothesis.
    """
    def __init__(
            self,
            pmm,
            minimizer,
            mean_n_sig_0,
            llhratio,
            **kwargs):
        r"""Creates a new ns-profile log-likelihood-ratio function with a
        null-hypothesis where :math:`n_{\mathrm{s}}` is fixed to
        ``mean_n_sig_0``.

        Parameters
        ----------
        pmm : instance of ParameterModelMapper
            The instance of ParameterModelMapper providing the mapping of
            global parameters to local parameters of individual models.
        minimizer : instance of Minimizer
            The Minimizer instance that should be used to minimize the negative
            of this log-likelihood ratio function.
        mean_n_sig_0 : float
            The fixed parameter value for the mean number of signal events of
            the null-hypothesis.
        llhratio : instance of MultiDatasetTCLLHRatio
            The instance of MultiDatasetTCLLHRatio, which should be used as
            log-likelihood function.
        """
        super().__init__(
            pmm=pmm,
            minimizer=minimizer,
            mean_n_sig_0=mean_n_sig_0,
            **kwargs)

        self.llhratio = llhratio

        if self._pmm.n_global_floating_params != 1:
            raise ValueError(
                'The log-likelihood-ratio function implemented by '
                f'{classname(self)} provides functionality only for LLH '
                'function with a single global fit parameter! Currently there '
                f'are {pmm.n_global_floating_params} global fit parameters '
                'defined!')

        # Define a member to hold the constant null-hypothesis log-likelihood
        # function value for ns=mean_n_sig_0.
        self._logL_0 = None

    @property
    def llhratio(self):
        """The instance of MultiDatasetTCLLHRatio, which should be used as
        log-likelihood function.
        """
        return self._llhratio

    @llhratio.setter
    def llhratio(self, obj):
        if not isinstance(obj, MultiDatasetTCLLHRatio):
            raise TypeError(
                'The llhratio property must be an instance of '
                'MultiDatasetTCLLHRatio! '
                f'Its current type is {classname(obj)}.')
        self._llhratio = obj

    def change_shg_mgr(
            self,
            shg_mgr):
        """Changes the source hypo group manager of all objects of this LLH
        ratio function, hence, calling the ``change_shg_mgr``
        method of the underlying MultiDatasetTCLLHRatio instance of this
        LLHRatio instance.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The new instance of SourceHypoGroupManager.
        """
        self._llhratio.change_shg_mgr(shg_mgr=shg_mgr)

    def initialize_for_new_trial(
            self,
            tl=None,
            **kwargs):
        """Initializes the log-likelihood-ratio function for a new trial.

        Parameters
        ----------
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used for timing
            measurements.
        """
        self._llhratio.initialize_for_new_trial(
            tl=tl,
            **kwargs)

        # Compute the constant log-likelihood function value for the
        # null-hypothesis.
        fitparam_values_0 = np.array([self._mean_n_sig_0], dtype=np.float64)
        (self._logL_0, grads_0) = self._llhratio.evaluate(
            fitparam_values=fitparam_values_0,
            tl=tl)

    def evaluate(
            self,
            fitparam_values,
            src_params_recarray=None,
            tl=None):
        """Evaluates the log-likelihood-ratio function and returns its value and
        global fit parameter gradients.

        Parameters
        ----------
        fitparam_values : instance of numpy ndarray
            The (1,)-shaped numpy 1D ndarray holding the current
            values of the global fit parameters.
            By definition of this LLH ratio function, it must contain the single
            fit parameter value for ns.
        src_params_recarray : instance of numpy record ndarray
            The numpy record ndarray of length N_sources holding the parameter
            names and values of all sources.
            See the documentation of the
            :meth:`skyllh.core.parameters.ParameterModelMapper.create_src_params_recarray`
            method for more information about this array.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used for timing
            measurements.

        Returns
        -------
        log_lambda : float
            The calculated log-lambda value of this log-likelihood-ratio
            function.
        grads : (1,)-shaped 1D ndarray
            The ndarray holding the gradient value of this log-likelihood-ratio
            for ns.
        """
        (logL, grads) = self._llhratio.evaluate(
            fitparam_values=fitparam_values,
            src_params_recarray=src_params_recarray,
            tl=tl)

        log_lambda = logL - self._logL_0

        return (log_lambda, grads)

    def calculate_ns_grad2(
            self,
            ns,
            ns_pidx,
            src_params_recarray,
            tl=None):
        """Calculates the second derivative w.r.t. ns of the log-likelihood
        ratio function.

        Parameters
        ----------
        ns : float
            The value of the global fit parameter ns.
        ns_pidx : int
            The index of the global fit parameter ns. By definition this must
            be ``0``.
        src_params_recarray : instance of numpy record ndarray
            The numpy record ndarray of length N_sources holding the parameter
            names and values of all sources.
            See the documentation of the
            :meth:`skyllh.core.parameters.ParameterModelMapper.create_src_params_recarray`
            method for more information about this array.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used for timing
            measurements.

        Returns
        -------
        nsgrad2 : float
            The second derivative w.r.t. ns of the log-likelihood ratio function
            for the given fit parameter values.
        """
        if ns_pidx != 0:
            raise ValueError(
                'The value of the ns_pidx argument must be 0! '
                f'Its current value is {ns_pidx}.')

        nsgrad2 = self._llhratio.calculate_ns_grad2(
            ns=ns,
            ns_pidx=ns_pidx,
            src_params_recarray=src_params_recarray,
            tl=tl)

        return nsgrad2
